import os
import yaml
import time
import logging
from typing import Optional, Union, Callable
import numpy as np
import argparse
import traceback
from typing import Optional
from sotodlib.utils.procs_pool import get_exec_env
import h5py
import copy
from sotodlib.coords import demod as demod_mm
from sotodlib.hwp import hwp_angle_model
from sotodlib import core
from sotodlib.preprocess import _Preprocess, Pipeline, processes
import sotodlib.preprocess.preprocess_util as pp_util
import sotodlib.site_pipeline.util as sp_util

logger = pp_util.init_logger("preprocess")

def multilayer_preprocess_tod(obs_id, 
                              configs_init,
                              configs_proc,
                              verbosity=0,
                              group_list=None, 
                              overwrite=False,
                              run_parallel=False):
    """Meant to be run as part of a batched script. Given a single
    Observation ID, this function uses an existing ManifestDb generated
    from a previous runof the processing pipeline, runs the pipeline using
    a second config, and outputs a new ManifestDb. Det groups must exist in
    the first dB to be included in the pipeline run on the second config.

    Arguments
    ----------
    obs_id: string or ResultSet entry
        obs_id or obs entry that is passed to context.get_obs
    configs_init: string or dictionary
        config file or loaded config directory for existing database
    configs_proc: string or dictionary
        config file or loaded config directory for processing database
        to be output
    group_list: None or list
        list of groups to run if you only want to run a partial update
    overwrite: bool
        if True, overwrite existing entries in ManifestDb
    verbosity: log level
        0 = error, 1 = warn, 2 = info, 3 = debug
    run_parallel: Bool
        If true preprocess_tod is called in a parallel process which returns
        dB info and errors and does no sqlite writing inside the function.
    """

    logger = pp_util.init_logger("preprocess", verbosity=verbosity)

    # list to hold error, destination file, and db data
    outputs_init = []
    outputs_proc = []

    if type(configs_init) == str:
        configs_init = yaml.safe_load(open(configs_init, "r"))
    context_init = core.Context(configs_init["context_file"])

    if type(configs_proc) == str:
        configs_proc = yaml.safe_load(open(configs_proc, "r"))
    context_proc = core.Context(configs_proc["context_file"])

    group_by_init, groups_init, error_init = pp_util.get_groups(obs_id, configs_init, context_init)
    group_by_proc, groups_proc, error_proc = pp_util.get_groups(obs_id, configs_proc, context_proc)

    if error_init is not None:
        if run_parallel:
            return error_init[0], [None, None], [None, None]
        else:
            return

    if error_proc is not None:
        if run_parallel:
            return error_proc[0], [None, None], [None, None]
        else:
            return

    if len(groups_init) > 0 and len(groups_proc) > 0:
        if (group_by_init != group_by_proc).any():
            raise ValueError('init and proc groups do not match')

    all_groups_proc = groups_proc.copy()
    for g in all_groups_proc:
        if g not in groups_init:
            groups_proc.remove(g)
            continue
        if group_list is not None:
            if g not in group_list:
                groups_proc.remove(g)
                continue
        if 'wafer.bandpass' in group_by_proc:
            if 'NC' in g:
                groups_proc.remove(g)
                continue
        try:
            meta = context_proc.get_meta(obs_id, dets = {gb:gg for gb, gg in zip(group_by_proc, g)})
        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: {obs_id} {g}\n{errmsg}\n{tb}")
            groups_proc.remove(g)
            continue

        if meta.dets.count == 0:
            groups_proc.remove(g)

    if len(groups_proc) == 0:
        logger.warning(f"group_list:{group_list} contains no overlap with "
                       f"groups in observation: {obs_id}:{all_groups_proc}. "
                       f"No analysis to run.")
        error = 'no_group_overlap'
        if run_parallel:
            return error, [None, None], [None, None]
        else:
            return

    if not(run_parallel):
        db_init = pp_util.get_preprocess_db(configs_init, group_by_proc)
        db_proc = pp_util.get_preprocess_db(configs_proc, group_by_proc)

    # pipeline for init config
    pipe_init = Pipeline(configs_init["process_pipe"],
                         plot_dir=configs_init["plot_dir"], logger=logger)
    # pipeline for processing config
    pipe_proc = Pipeline(configs_proc["process_pipe"],
                         plot_dir=configs_proc["plot_dir"], logger=logger)

    if configs_proc.get("lmsi_config", None) is not None:
        make_lmsi = True
    else:
        make_lmsi = False

    # loop through and reduce each group
    n_fail = 0
    for group in groups_proc:
        logger.info(f"Beginning run for {obs_id}:{group}")
        dets = {gb:gg for gb, gg in zip(group_by_proc, group)}
        try:
            error, outputs_grp_init, _, aman = pp_util.preproc_or_load_group(obs_id, configs_init,
                                                                             dets=dets, logger=logger)
            if error is None:
                outputs_init.append(outputs_grp_init)

            init_fields = aman.preprocess._fields.copy()
            init_fields.pop('valid_data', None)

            outputs_grp_proc = pp_util.save_group(obs_id, configs_proc, dets,
                                      context_proc, subdir='temp_proc')

            # tags from context proc
            tags_proc = np.array(context_proc.obsdb.get(aman.obs_info.obs_id, tags=True)['tags'])
            if "tags" in aman._fields:
                aman.move("tags", None)
            aman.wrap('tags', tags_proc)

            # now run the pipeline on the processed axis manager
            logger.info(f"Beginning processing pipeline for {obs_id}:{group}")
            proc_aman, success = pipe_proc.run(aman)
            proc_aman.wrap('pcfg_ref', pp_util.get_pcfg_check_aman(pipe_init))

            # remove fields found in aman.preprocess from proc_aman
            for fld_init in init_fields:
                if fld_init in proc_aman:
                    proc_aman.move(fld_init, None)

        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: {obs_id} {group}\n{errmsg}\n{tb}")
            n_fail += 1
            continue
        if success != 'end':
            # If a single group fails we don't log anywhere just mis an entry in the db.
            logger.info(f"ERROR: {obs_id} {group}\nFailed at step {success}") 
            n_fail += 1
            continue

        logger.info(f"Saving data to {outputs_grp_proc['temp_file']}:{outputs_grp_proc['db_data']['dataset']}")
        proc_aman.save(outputs_grp_proc['temp_file'], outputs_grp_proc['db_data']['dataset'], overwrite)

        if run_parallel:
            outputs_proc.append(outputs_grp_proc)
        else:
            logger.info(f"Saving to database under {outputs_grp_proc['db_data']}")
            if len(db_proc.inspect(outputs_grp_proc['db_data'])) == 0:
                h5_path = os.path.relpath(outputs_grp_proc['temp_file'],
                        start=os.path.dirname(configs_proc['archive']['index']))
                db_proc.add_entry(outputs_grp_proc['db_data'], h5_path)

    if make_lmsi:
        from pathlib import Path
        import lmsi.core as lmsi

        if os.path.exists(new_plots):
            lmsi.core([Path(x.name) for x in Path(new_plots).glob("*.png")],
                      Path(configs1["lmsi_config"]),
                      Path(os.path.join(new_plots, 'index.html')))

    if run_parallel:
        if n_fail == len(groups_proc):
            # If no groups make it to the end of the processing return error.
            logger.info(f'ERROR: all groups failed for {obs_id}')
            error = 'all_fail'
            return error, [obs_id, 'all groups'], [obs_id, 'all groups']
        else:
            logger.info('Returning data to futures')
            error = None
            return error, outputs_init, outputs_proc


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('configs_init', help="Preprocessing Configuration File for existing database")
    parser.add_argument('configs_proc', help="Preprocessing Configuration File for new database")
    parser.add_argument(
        '--query', 
        help="Query to pass to the observation list. Use \\'string\\' to "
             "pass in strings within the query.",  
        type=str
    )
    parser.add_argument(
        '--obs-id',
        help="obs-id of particular observation if we want to run on just one"
    )
    parser.add_argument(
        '--overwrite',
        help="If true, overwrites existing entries in the database",
        action='store_true',
    )
    parser.add_argument(
        '--min-ctime',
        help="Minimum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--max-ctime',
        help="Maximum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--update-delay',
        help="Number of days (unit is days) in the past to start observation list.",
        type=int
    )
    parser.add_argument(
        '--tags',
        help="Observation tags. Ex: --tags 'jupiter' 'setting'",
        nargs='*',
        type=str
    )
    parser.add_argument(
        '--planet-obs',
        help="If true, takes all planet tags as logical OR and adjusts related configs",
        action='store_true',
    )
    parser.add_argument(
        '--verbosity',
        help="increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug",
        default=2,
        type=int
    )
    parser.add_argument(
        '--nproc',
        help="Number of parallel processes to run on.",
        type=int,
        default=4
    )
    parser.add_argument(
        '--raise-error',
        help="Raise an error upon completion if any obsids or groups fail.",
        type=bool,
        default=False
    )
    return parser

def _main(executor: Union["MPICommExecutor", "ProcessPoolExecutor"],
          as_completed_callable: Callable,
          configs_init: str,
          configs_proc: str,
          query: Optional[str] = None,
          obs_id: Optional[str] = None,
          overwrite: bool = False,
          min_ctime: Optional[int] = None,
          max_ctime: Optional[int] = None,
          update_delay: Optional[int] = None,
          tags: Optional[str] = None,
          planet_obs: bool = False,
          verbosity: Optional[int] = None,
          nproc: Optional[int] = 4,
          raise_error: Optional[bool] = False):

    logger = pp_util.init_logger("preprocess", verbosity=verbosity)

    configs_init, context_init = pp_util.get_preprocess_context(configs_init)
    configs_proc, context_proc = pp_util.get_preprocess_context(configs_proc)

    errlog = os.path.join(os.path.dirname(configs_proc['archive']['index']),
                          'errlog.txt')

    obs_list = sp_util.get_obslist(context_proc, query=query, obs_id=obs_id, min_ctime=min_ctime,
                                   max_ctime=max_ctime, update_delay=update_delay, tags=tags,
                                   planet_obs=planet_obs)
    if len(obs_list)==0:
        logger.warning(f"No observations returned from query: {query}")

    # clean up lingering files from previous incomplete runs
    policy_dir_init = os.path.join(os.path.dirname(configs_init['archive']['policy']['filename']), 'temp')
    policy_dir_proc = os.path.join(os.path.dirname(configs_proc['archive']['policy']['filename']), 'temp_proc')
    for obs in obs_list:
        obs_id = obs['obs_id']
        pp_util.cleanup_obs(obs_id, policy_dir_init, errlog, configs_init, context_init,
                            subdir='temp', remove=overwrite)
        pp_util.cleanup_obs(obs_id, policy_dir_proc, errlog, configs_proc, context_proc,
                            subdir='temp_proc', remove=overwrite)

    # remove datasets from final archive file not found in init db
    pp_util.cleanup_archive(configs_init, logger)
    pp_util.cleanup_archive(configs_proc, logger)

    run_list = []

    if overwrite or not os.path.exists(configs_proc['archive']['index']):
        # run on all if database doesn't exist
        for obs in obs_list:
            #run on all if database doesn't exist
            run_list = [ (o,None) for o in obs_list]
            group_by_proc = np.atleast_1d(configs_proc['subobs'].get('use', 'detset'))
    else:
        db = core.metadata.ManifestDb(configs_proc['archive']['index'])

        for obs in obs_list:
            x = db.inspect({'obs:obs_id': obs["obs_id"]})
            if x is None or len(x) == 0:
                run_list.append( (obs, None) )
            else:
                group_by_proc, groups_proc, _ = pp_util.get_groups(obs["obs_id"], configs_proc, context_proc)
                if len(x) != len(groups_proc):
                    [groups_proc.remove([a[f'dets:{gb}'] for gb in group_by_proc]) for a in x]
                    run_list.append( (obs, groups_proc) )

    logger.info(f'Run list created with {len(run_list)} obsids')

    n_fail = 0

    # run write_block obs-ids in parallel at once then write all to the sqlite db.
    futures = [executor.submit(multilayer_preprocess_tod, obs_id=r[0]['obs_id'],
                group_list=r[1], verbosity=verbosity,
                configs_init=configs_init,
                configs_proc=configs_proc,
                overwrite=overwrite, run_parallel=True) for r in run_list]
    for future in as_completed_callable(futures):
        logger.info('New future as_completed result')
        try:
            err, db_datasets_init, db_datasets_proc = future.result()
            if err is not None:
                n_fail += 1
        except Exception as e:
            errmsg = f'{type(e)}: {e}'
            tb = ''.join(traceback.format_tb(e.__traceback__))
            logger.info(f"ERROR: future.result()\n{errmsg}\n{tb}")
            f = open(errlog, 'a')
            f.write(f'\n{time.time()}, future.result() error\n{errmsg}\n{tb}\n')
            f.close()
            n_fail += 1
            continue
        futures.remove(future)

        if db_datasets_init:
            if err is None:
                for db_dataset in db_datasets_init:
                    logger.info(f'Processing future result db_dataset: {db_datasets_init}')
                    pp_util.cleanup_mandb(err, db_dataset, configs_init, logger, overwrite)
            else:
                pp_util.cleanup_mandb(err, db_datasets_init, configs_init, logger, overwrite)

        if db_datasets_proc:
            if err is None:
                logger.info(f'Processing future dependent result db_dataset: {db_datasets_proc}')
                for db_dataset in db_datasets_proc:
                    pp_util.cleanup_mandb(err, db_dataset, configs_proc, logger, overwrite)
            else:
                pp_util.cleanup_mandb(err, db_datasets_proc, configs_proc, logger, overwrite)

    if raise_error and n_fail > 0:
        raise RuntimeError(f"multilayer_preprocess_tod: {n_fail}/{len(run_list)} obs_ids failed")


def main(configs_init: str,
         configs_proc: str,
         query: Optional[str] = None,
         obs_id: Optional[str] = None,
         overwrite: bool = False,
         min_ctime: Optional[int] = None,
         max_ctime: Optional[int] = None,
         update_delay: Optional[int] = None,
         tags: Optional[str] = None,
         planet_obs: bool = False,
         verbosity: Optional[int] = None,
         nproc: Optional[int] = 4,
         raise_error: Optional[bool] = False):

    rank, executor, as_completed_callable = get_exec_env(nproc)
    if rank == 0:
        _main(executor=executor,
              as_completed_callable=as_completed_callable,
              configs_init=configs_init,
              configs_proc=configs_proc,
              query=query,
              obs_id=obs_id,
              overwrite=overwrite,
              min_ctime=min_ctime,
              max_ctime=max_ctime,
              update_delay=update_delay,
              tags=tags,
              planet_obs=planet_obs,
              verbosity=verbosity,
              nproc=nproc,
              raise_error=raise_error)


if __name__ == '__main__':
    sp_util.main_launcher(main, get_parser)



# Copyright (c) 2018-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import h5py
import os
import pickle

from astropy import units as u
import ephem
import healpy as hp
import numpy as np
from scipy.constants import h, c, k
from scipy.interpolate import RectBivariateSpline
from scipy.signal import fftconvolve
import toml
import traitlets

import toast
from toast.timing import function_timer, Timer
from toast import qarray as qa
from toast.data import Data
from toast.traits import (
    trait_docs, Int, Unicode, Unit, Bool, Quantity, Float, Instance
)
from toast.ops.operator import Operator
from toast.utils import Environment, Logger
from toast.observation import default_values as defaults
from toast.coordinates import azel_to_radec

from . import utils


def to_JD(t):
    # Unix time stamp to Julian date
    # (days since -4712-01-01 12:00:00 UTC)
    return t / 86400.0 + 2440587.5


def to_MJD(t):
    # Convert Unix time stamp to modified Julian date
    # (days since 1858-11-17 00:00:00 UTC)
    return to_JD(t) - 2400000.5


def to_DJD(t):
    # Convert Unix time stamp to Dublin Julian date
    # (days since 1899-12-31 12:00:00)
    # This is the time format used by PyEphem
    return to_JD(t) - 2415020


@trait_docs
class SimCatalog(Operator):
    """Operator that generates variable and static point source signal.

    Signal is generated by sampling the provided beam map at appropriate
    locations and scaling the resulting signal to match the perceived
    intensity of the source.

    Source SED is convolved with the detector bandpass recorded in the
    focalplane table.

    Example catalog entries:

    .. highlight:: toml
    .. code-block:: toml

        [example_static_source]
        # Celestial coordinate are always given in degrees
        ra_deg = 30
        dec_deg = -30
        # the SED can be specified using an arbitrary number of
        # frequency bins.  The SED is interpolated in log-log space to
        # convolve with the detector bandpass
        # Use either `flux_density_mJy` or `flux_density_Jy` and adjust
        # the values accordingly
        freqs_ghz = [ 1.0, 1000.0,]
        flux_density_mJy = [ 10.0, 1.0,]
        # Omitting polarization fraction results in an
        # unpolarized source
        pol_frac = 0.1
        pol_angle_deg = 0

        [example_variable_source]
        ra_deg = 30
        dec_deg = -25
        freqs_ghz = [ 1.0, 1000.0,]
        # An arbitrary number of SED vectors can be provided but the
        # location of the frequency bins is fixed.  Effective SED is
        # interpolated between the specified epochs.
        flux_density_Jy = [ [ 10.0, 1.0,], [ 30.0, 10.0,], [ 10.0, 1.0,],]
        # Omitting the times_mjd entry resuls in a static source
        times_mjd = [ 59000.0, 60000.0, 61000.0,]
        # The polarization properties can also vary
        pol_frac = [ 0.05, 0.15, 0.05,]
        pol_angle_deg = [ 45, 45, 45,]

        [example_transient_source]
        ra_deg = 30
        dec_deg = -20
        freqs_ghz = [ 1.0, 1000.0,]
        flux_density_Jy = [ [ 10.0, 1.0,], [ 30.0, 10.0,],]
        # Difference between a variable and transient source is
        # simply that the specified epochs do not cover the entire
        # simulation time span.  The operator will not extrapolate
        # outside the epochs.
        times_mjd = [ 59410.0, 59411.0,]
    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    times = Unicode(
        defaults.times,
        help="Observation shared key for timestamps",
    )

    hwp_angle = Unicode(
        defaults.hwp_angle,
        help="Observation shared key for HWP angle",
    )

    catalog_file = Unicode(
        None,
        allow_none=True,
        help="Name of the TOML catalog file",
    )

    beam_file = Unicode(
        None,
        allow_none=True,
        help="HDF5 file that stores the simulated beam",
    )

    det_data = Unicode(
        defaults.det_data,
        help="Observation detdata key for simulated signal",
    )

    det_data_units = Unit(
        defaults.det_data_units, help="Output units if creating detector data"
    )

    detector_pointing = Instance(
        klass=Operator,
        allow_none=True,
        help="Operator that translates boresight Az/El pointing into detector frame",
    )

    @traitlets.validate("catalog_file")
    def _check_catalog_file(self, proposal):
        filename = proposal["value"]
        if filename is not None and not os.path.isfile(filename):
            raise traitlets.TraitError(f"Catalog file does not exist: {filename}")
        return filename

    @traitlets.validate("beam_file")
    def _check_beam_file(self, proposal):
        beam_file = proposal["value"]
        if beam_file is not None and not os.path.isfile(beam_file):
            raise traitlets.TraitError(f"{beam_file} is not a valid beam file")
        return beam_file

    @traitlets.validate("detector_pointing")
    def _check_detector_pointing(self, proposal):
        detpointing = proposal["value"]
        if detpointing is not None:
            if not isinstance(detpointing, Operator):
                raise traitlets.TraitError(
                    "detector_pointing should be an Operator instance"
                )
            # Check that this operator has the traits we expect
            for trt in [
                "view",
                "boresight",
                "shared_flags",
                "shared_flag_mask",
                "quats",
                "coord_in",
                "coord_out",
            ]:
                if not detpointing.has_trait(trt):
                    msg = f"detector_pointing operator should have a '{trt}' trait"
                    raise traitlets.TraitError(msg)
        return detpointing

    @function_timer
    def _load_catalog(self):
        # Load the TOML into a dictionary
        with open(self.catalog_file, "r") as f:
            self.catalog = toml.loads(f.read())
        # Translate each source position into a vector for rapid
        # distance calculations
        for source_name, source_dict in self.catalog.items():
            lon = source_dict["ra_deg"]
            lat = source_dict["dec_deg"]
            source_dict["vec"] = hp.dir2vec(lon, lat, lonlat=True)
        return

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        # Store of per-detector beam properties.  Eventually we could modify the
        # operator traits to list files per detector, per wafer, per tube, etc.
        # For now, we use the same beam for all detectors, so this will have only
        # one entry.
        self.beam_props = dict()

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()
        comm = data.comm

        for trait in "catalog_file", "beam_file", "detector_pointing":
            value = getattr(self, trait)
            if value is None:
                msg = f"You must set `{trait}` before running SimCatalog"
                raise RuntimeError(msg)

        self._load_catalog()

        for obs in data.obs:
            prefix = f"{comm.group} : {obs.name}"

            # Make sure detector data output exists.  If not, create it
            # with units of Kelvin.

            dets = obs.select_local_detectors(detectors)
            exists = obs.detdata.ensure(
                self.det_data, detectors=dets, create_units=self.det_data_units
            )
            det_units = obs.detdata[self.det_data].units
            scale = toast.utils.unit_conversion(u.K, det_units)

            self._observe_catalog(
                data,
                obs,
                prefix,
                dets,
                scale,
            )

        return

    @function_timer
    def _get_beam_map(self, det):
        """
        Construct a 2-dimensional interpolator for the beam
        """
        # Read in the simulated beam.  We could add operator traits to
        # specify whether to load different beams based on detector,
        # wafer, tube, etc and check that key here.
        log = Logger.get()
        if "ALL" in self.beam_props:
            # We have already read the single beam file.
            beam_dic = self.beam_props["ALL"]
        else:
            with h5py.File(self.beam_file, 'r') as f_t:
                beam_dic = {}
                beam_dic["data"] = f_t["beam"][:]
                beam_dic["size"] = [[f_t["beam"].attrs["size"], f_t["beam"].attrs["res"]], [f_t["beam"].attrs["npix"], 1]]
                self.beam_props["ALL"] = beam_dic
        description = beam_dic["size"]  # 2d array [[size, res], [n, 1]]
        model = beam_dic["data"].copy()
        model /= np.amax(model)
        res = description[0][1] * u.degree
        beam_solid_angle = np.sum(model) * res**2

        # DEBUG begin
        # These commands add a tail to the beam that points towards the horizon
        #nx, ny = np.shape(model)
        #nhalf = nx // 2
        #w = 10
        #model[nhalf - w : nhalf + w + 1, 0 : nhalf] = 1
        # DEBUG end

        n = int(description[1][0])
        size = description[0][0] * u.degree
        w = size.to_value(u.rad) / 2
        x = np.linspace(-w, w, n)
        y = np.linspace(-w, w, n)
        beam = RectBivariateSpline(x, y, model)
        r = np.sqrt(w**2 + w**2)
        return beam, r, beam_solid_angle

    @function_timer
    def _observe_catalog(
        self,
        data,
        obs,
        prefix,
        dets,
        scale,
    ):
        """
        Observe the catalog with each detector in tod
        """
        log = Logger.get()

        # Get a view of the data which contains just this single
        # observation
        obs_data = data.select(obs_name=obs.name)
        focalplane = obs.telescope.focalplane

        times_mjd = to_MJD(obs.shared[self.times].data)
        if self.hwp_angle in obs.shared:
            hwp_angle = obs.shared[self.hwp_angle].data
        else:
            hwp_angle = None
        beam = None

        for idet, det in enumerate(dets):
            bandpass = obs.telescope.focalplane.bandpass
            signal = obs.detdata[self.det_data][det]

            self.detector_pointing.apply(obs_data, detectors=[det])
            det_quat = obs_data.obs[0].detdata[self.detector_pointing.quats][det]

            # Convert Az/El quaternion of the detector into angles
            # `psi` includes the rotation to the detector polarization
            #sensitive direction

            det_theta, det_phi, det_psi = qa.to_iso_angles(det_quat)
            det_vec = hp.dir2vec(det_theta, det_phi).T.copy()
            det_psi_pol = focalplane[det]["pol_ang"]

            if beam is None or not "ALL" in self.beam_props:
                beam, beam_radius, beam_solid_angle = self._get_beam_map(det)
            dp_radius = np.cos(beam_radius)

            for source_name, source_dict in self.catalog.items():
                # Is this source close enough to register?
                dp = np.dot(det_vec, source_dict["vec"])
                hit = dp > dp_radius
                nhit = np.sum(hit)
                if nhit == 0:
                    continue

                # Get the appropriate source SED and convolve with the
                # detector bandpass
                if "times_mjd" in source_dict:
                    source_times = np.array(source_dict["times_mjd"])
                    ind = np.array(np.searchsorted(source_times, times_mjd))
                    # When time stamps fall outside the period covered by
                    # source time, we assume the source went quiet
                    good = np.logical_and(ind > 0, ind < len(source_times))
                    hit *= good
                    nhit = np.sum(hit)
                    if nhit == 0:
                        # This source is not active during our observation
                        continue
                    ind = ind[hit]
                    lengths = source_times[ind] - source_times[ind - 1]
                    right_weights = (source_times[ind] - times_mjd[hit]) \
                                    / lengths
                    left_weights = 1 - right_weights
                    # useful shorthands
                    freq = np.array(source_dict["freqs_ghz"]) * u.GHz
                    if "flux_density_Jy" in source_dict:
                        seds = np.array(source_dict["flux_density_Jy"]) * u.Jy
                    elif "flux_density_mJy" in source_dict:
                        seds = np.array(source_dict["flux_density_mJy"]) * u.mJy
                    else:
                        msg = f"No flux density for {source_name}"
                        raise RuntimeError(msg)
                    # Mean SED used for bandpass convolution
                    wright = np.mean(right_weights)
                    wleft = 1 - wright
                    cindex = int(np.median(ind))
                    sed_mean = wleft * seds[cindex - 1] + wright * seds[cindex]
                    # Time-dependent amplitude to scale the mean SED
                    cfreq = bandpass.center_frequency(det, alpha=-1)
                    amplitudes = []
                    for sed in seds:
                        # Interpolate the SED to the detector central frequency
                        # in log-log domain where power-law spectra are
                        # linear
                        amp = np.exp(np.interp(
                                np.log(cfreq.to_value(u.GHz)),
                                np.log(freq.to_value(u.GHz)),
                                np.log(sed.to_value(u.Jy))
                        ))
                        amplitudes.append(amp)
                    amplitudes = np.array(amplitudes)
                    # This is the time-dependent amplitude relative to
                    # sed_mean
                    amplitude = (
                        left_weights * amplitudes[ind - 1] +
                        right_weights * amplitudes[ind]
                    )
                    amplitude /=  (
                        wleft * amplitudes[cindex - 1] +
                        wright * amplitudes[cindex]
                    )
                    if "pol_frac" in source_dict:
                        pol_fracs = np.array(source_dict["pol_frac"])
                        pol_frac = (
                            left_weights * pol_fracs[ind - 1] +
                            right_weights * pol_fracs[ind]
                        )
                        pol_angles = np.unwrap(
                            np.radians(source_dict["pol_angle_deg"])
                        )
                        pol_angle = np.array(
                            left_weights * pol_angles[ind - 1] +
                            right_weights * pol_angles[ind]
                        )
                    else:
                        pol_frac = None
                        pol_angle = None
                else:
                    freq = np.array(source_dict["freqs_ghz"]) * u.GHz
                    if "flux_density_Jy" in source_dict:
                        sed_mean = np.array(source_dict["flux_density_Jy"]) * u.Jy
                    elif "flux_density_mJy" in source_dict:
                        sed_mean = np.array(source_dict["flux_density_mJy"]) * u.mJy
                    else:
                        msg = f"No flux density for {source_name}"
                        raise RuntimeError(msg)
                    if "pol_frac" in source_dict:
                        pol_frac = np.array(source_dict["pol_frac"])
                        pol_angle = np.radians(source_dict["pol_angle_deg"])
                    else:
                        pol_frac = None
                        pol_angle = None
                    amplitude = 1
                    
                # Convolve the SED with the detector bandpass
                flux_density = bandpass.convolve(
                    det,
                    freq,
                    sed_mean.to_value(u.Jy),
                )

                # Convert the flux density to peak temperature
                temperature = (
                    flux_density
                    / beam_solid_angle.to_value(u.rad**2)
                    / bandpass.kcmb2jysr(det)
                )

                # Modulate the temperature in time
                temperature = temperature * amplitude

                # modulate temperature by polarization
                if pol_frac is not None:
                    Q = temperature * pol_frac * np.cos(2 * pol_angle)
                    U = temperature * pol_frac * np.sin(2 * pol_angle)
                    psi = det_psi[hit]
                    if hwp_angle is not None:
                        psi +=  2 * hwp_angle[hit]
                    temperature += Q * np.cos(2 * psi) + U * np.sin(2 * psi)

                # Interpolate the beam map at appropriate locations
                source_theta = np.radians(90 - source_dict["dec_deg"])
                source_phi = np.radians(source_dict["ra_deg"])
                phi_diff = (det_phi[hit] - source_phi + np.pi) % (2 * np.pi) - np.pi
                x = phi_diff * np.cos(np.pi / 2 - det_theta[hit])
                y = det_theta[hit] - source_theta
                # Rotate into the beam frame
                psi = det_psi[hit] - det_psi_pol.to_value(u.rad)
                x_beam = np.cos(psi) * x - np.sin(psi) * y
                y_beam = np.sin(psi) * x + np.cos(psi) * y
                sig = beam(x_beam, y_beam, grid=False) * temperature
                signal[hit] += scale * sig

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = {
            "shared": [
                self.times,
            ],
        }
        return req

    def _provides(self):
        return {
            "detdata": [
                self.det_data,
            ]
        }

    def _accelerators(self):
        return list()



"""check_book.py

This module an entry point to io.check_book, for checking obs/oper
Books for internal consistency & proper schema.  It may also be used
to create/update an ObsFileDb for such Books.

A configuration file can be used to set the ObsFileDb filename and
root path for ObsFileDb entries.

The config file can also be used to enable work-arounds and bypass
certain exceptions (which should not be necessary on compliant books.)

At the of this writing a minimal config file might be simply:

.. code-block:: yaml

    # Database setup (this is the default).
    obsfiledb: './obsfiledb.sqlite'

    # For obsdb filenames, path relative to which those names should
    # be specified.  (/ is the default.)
    root_path: '/'

    # Work-arounds
    extra_extra_files: ['frame_splits.txt']


But here is a more complete example, with lots of work-arounds
enabled:

.. code-block:: yaml

    # Database setup (this is the default).
    obsfiledb: './obsfiledb.sqlite'

    # For obsdb filenames, path relative to which those names should
    # be specified.  (/ is the default.)
    root_path: '/'

    # Work-arounds
    stream_file_pattern: 'D_obs_{stream_id}_{index:03d}.g3'
    extra_extra_files: ['frame_splits.txt']
    sample_range_inclusive_hack: True
    tolerate_missing_ancil: True
    tolerate_missing_ancil_timestamps: True
    tolerate_timestamps_value_discrepancy: False

    # Tolerate arbitrary extra files, except explicitly named ones
    tolerate_stray_files: True
    banned_files: ['frame_splits.txt']

    # If stream_ids are not provided in metadata, list them here.
    stream_ids:
      ufm_mv14
      ufm_mv18
      ufm_mv19
      ufm_mv22
      ufm_mv6 
      ufm_mv7 
      ufm_mv9 

    # If detset names are not provided in metadata, provide a map from
    # stream_id to detset name here.
    detset_map:
      ufm_mv14: sch_mv14
      ufm_mv18: sch_mv18
      ufm_mv19: sch_mv19
      ufm_mv22: sch_mv22
      ufm_mv6:  sch_mv6
      ufm_mv7:  sch_mv7
      ufm_mv9:  sch_mv9

"""

import sotodlib
from sotodlib.core import metadata
from sotodlib.io import check_book

import argparse
import sys
import yaml

from . import util


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.description = """

        Scan an "obs" or "oper" book and check for schema compliance;
        optionally update an obsfiledb.

    """
    parser.add_argument('book_dir',
                        help="Path to the Book.")
    parser.add_argument('--config', '-c',
                        help="Path to config file with work-arounds and ObsFileDb config.")
    parser.add_argument('--add', action='store_true',
                        help="After inspecting the book, add it to the ObsFileDb.")
    parser.add_argument('--overwrite', action='store_true',
                        help="If adding to ObsFileDb, remove existing references "
                        "to this obs first (prevents 'UNIQUE constraint' error).")
    parser.add_argument('--verbose', '-v', action='store_true')
    return parser


def scan_book_dir(book_dir, logger, config, prep_obsfiledb=False):
    """Run the BookScanner on book_dir.

    Returns:
      ok : bool
        True only if the book passed the checks.
      obsfiledb_info : dict or None
        If prep_obsfiledb, and ok, then this dict has the info needed
        for updating obsfiledb (see add_to_obsfiledb).

    (Note this function is used by update-obsdb, as well.)
    """
    logger.info(f'Examining {book_dir}')
    bs = check_book.BookScanner(book_dir, config, logger=logger)

    bs.go()
    bs.report()

    if len(bs.results['errors']):
        logger.error('Cannot register this obs due to errors.')
        return False, None

    if prep_obsfiledb:
        detset_rows, file_rows = bs.get_obsfiledb_info(config.get('root_path', '/'))
        return True, {
            'detset_rows': detset_rows,
            'file_rows': file_rows,
        }
    return True, None

def add_to_obsfiledb(info, logger, config, overwrite=False):
    """Add observation info to the obsfiledb specified in config.  The
    "info" dict is the one returned by scan_book_dir.  If overwrite,
    then file entries in the obsfiledb, for this obs_id, will be
    replaced.

    (Note this function is used by update-obsdb, as well.)
    """
    # Write to obsfiledb
    obsfiledb_file = config.get('obsfiledb', 'obsfiledb.sqlite')
    logger.debug('Updating %s ...' % obsfiledb_file)
    db = metadata.ObsFileDb(obsfiledb_file)

    if overwrite:
        # Note this only drops the obs ... if detsets need to be
        # rewritten, you'd better start over entirely.
        logger.debug(' -- removing any existing references.')
        db.drop_obs(info['file_rows'][0]['obs_id'])

    logger.debug(
        ' -- adding %i detsets and %i file refs' % (len(info['detset_rows']), 
        len(info['file_rows']))
    )
    for name, dets in info['detset_rows']:
        dets_already = set(db.get_dets(name))
        if len(dets_already) == 0:
            logger.debug(f' -- new detset: {name}')
        new_dets = [d for d in dets if d not in dets_already]
        if len(new_dets) > 0:
            if len(dets_already):
                logger.debug(f' -- extending det count by {len(new_dets)}')
            db.add_detset(name, new_dets)
    for row in info['file_rows']:
        db.add_obsfile(**row)


def main(book_dir, config=None, add=None, overwrite=None, verbose=None):
    logger = util.init_logger(__name__, 'check_book: ',
                              verbosity=(3 if verbose else 2))

    if config is not None:
        logger.debug(f'Loading config from {config}')
        config = yaml.safe_load(open(config, 'rb'))
    else:
        config = {}

    ok, info = scan_book_dir(book_dir, logger, config, prep_obsfiledb=add)
    if not ok:
        logger.error('Cannot register this obs due to errors.')
        sys.exit(1)

    if add:
        add_to_obsfiledb(info, logger, config, overwrite=overwrite)


if __name__ == '__main__':
    util.main_launcher(main, get_parser)



"""
Various AxisManager utility functions
"""

import numpy as np
from sotodlib.core import AxisManager

class RestrictionException(Exception):
    """Exception for when cannot restrict AxisManager properly"""

def restrict_to_times(am, t0, t1, in_place=False):
    """
    Restricts axis manager to a time range (t0, t1)
    """
    m = (t0 <= am.timestamps) & (am.timestamps < t1)
    if not m.any():
        raise RestrictionException()
    i0, i1 = np.where(m)[0][[0, -1]] + am.samps.offset
    return am.restrict('samps', (i0, i1+1), in_place=in_place)

def dict_to_am(d, skip_bad_types=False):
    """
    Attempts to convert a dictionary into an AxisManager. This can be used on
    dicts containing basic types such as (str, int, float) along with numpy
    arrays. The AxisManager will not have any structure such as "axes", but
    this is useful if you want to nest semi-arbitrary data such as the "meta"
    dict into an AxisManager.

    Args
    -----
    d : dict
        Dict to convert ot axismanager
    skip_bad_types : bool
        If True, will skip any value that is not a str, int, float, or
        np.ndarray. If False, this will raise an error for invalid types.
    """
    allowed_types = (str, int, float, np.ndarray)
    am = AxisManager()
    for k, v in d.items():
        if isinstance(v, dict):
            am.wrap(k, dict_to_am(v))
        elif isinstance(v, allowed_types):
            am.wrap(k, v)
        elif not skip_bad_types:
            raise ValueError(
                f"Key {k} is of type {type(v)} which cannot be wrapped by an "
                 "axismanager")
    return am



from typing import Optional, Any, Union
from sqlalchemy import create_engine, exc
from sqlalchemy.orm import declarative_base, Mapped, mapped_column, sessionmaker
import importlib
import numpy as np
import so3g
from pixell import enmap, fft, resample, tilemap, bunch, utils as putils

from .. import coords, core, tod_ops


def deslope_el(tod, el, srate, inplace=False):
    if not inplace: tod = tod.copy()
    putils.deslope(tod, w=1, inplace=True)
    f     = fft.rfftfreq(tod.shape[-1], 1/srate)
    fknee = 3.0
    with putils.nowarn():
        iN = (1+(f/fknee)**-3.5)**-1
    b  = 1/np.sin(el)
    putils.deslope(b, w=1, inplace=True)
    Nb = fft.irfft(iN*fft.rfft(b),b*0)
    amp= np.sum(tod*Nb,-1)/np.sum(b*Nb)
    tod-= amp[:,None]*b
    return tod

class ArrayZipper:
    def __init__(self, shape, dtype, comm=None):
        self.shape = shape
        self.ndof  = int(np.prod(shape))
        self.dtype = dtype
        self.comm  = comm

    def zip(self, arr):  return arr.reshape(-1)

    def unzip(self, x):  return x.reshape(self.shape).astype(self.dtype, copy=False)

    def dot(self, a, b):
        return np.sum(a*b) if self.comm is None else self.comm.allreduce(np.sum(a*b))

class MapZipper:
    def __init__(self, shape, wcs, dtype, comm=None):
        self.shape, self.wcs = shape, wcs
        self.ndof  = int(np.prod(shape))
        self.dtype = dtype
        self.comm  = comm

    def zip(self, map): return np.asarray(map.reshape(-1))

    def unzip(self, x): return enmap.ndmap(x.reshape(self.shape), self.wcs).astype(self.dtype, copy=False)

    def dot(self, a, b):
        return np.sum(a*b) if self.comm is None else putils.allreduce(np.sum(a*b), self.comm)

class TileMapZipper:
    def __init__(self, geo, dtype, comm):
        self.geo   = geo
        self.comm  = comm
        self.dtype = dtype
        self.ndof  = geo.size

    def zip(self, map):
        return np.asarray(map.reshape(-1))

    def unzip(self, x):
        return tilemap.TileMap(x.reshape(self.geo.pre+(-1,)).astype(self.dtype, copy=False), self.geo)

    def dot(self, a, b):
        return np.sum(a*b) if self.comm is None else putils.allreduce(np.sum(a*b), self.comm)

class MultiZipper:
    def __init__(self):
        self.zippers = []
        self.ndof    = 0
        self.bins    = []

    def add(self, zipper):
        self.zippers.append(zipper)
        self.bins.append([self.ndof, self.ndof+zipper.ndof])
        self.ndof += zipper.ndof

    def zip(self, *objs):
        return np.concatenate([zipper.zip(obj) for zipper, obj in zip(self.zippers, objs)])

    def unzip(self, x):
        res = []
        for zipper, (b1,b2) in zip(self.zippers, self.bins):
            res.append(zipper.unzip(x[b1:b2]))
        return res

    def dot(self, a, b):
        res = 0
        for (b1,b2), dof in zip(self.bins, self.zippers):
            res += dof.dot(a[b1:b2],b[b1:b2])
        return res

def inject_map(obs, map, recenter=None, interpol=None):
    # Infer the stokes components
    map = map.preflat
    if map.shape[0] not in [1,2,3]:
        raise ValueError("Map to inject must have either 1, 2 or 3 components, corresponding to T, QU and TQU.")
    comps = infer_comps(map.shape[0])
    # Support recentering the coordinate system
    if recenter is not None:
        ctime  = obs.timestamps
        rot    = recentering_to_quat_lonlat(*evaluate_recentering(recenter, ctime=ctime[len(ctime)//2], geom=(map.shape, map.wcs), site=unarr(obs.site)))
    else: rot = None
    # Set up our pointing matrix for the map
    pmat  = coords.pmat.P.for_tod(obs, comps=comps, geom=(map.shape, map.wcs), rot=rot, threads="domdir", interpol=interpol)
    # And perform the actual injection
    pmat.from_map(map.extract(map.shape, map.wcs), dest=obs.signal)

def safe_invert_div(div, lim=1e-2, lim0=np.finfo(np.float32).tiny**0.5):
    try:
        # try setting up a context manager that limits the number of threads
        from threadpoolctl import threadpool_limits
        cm = threadpool_limits(limits=1, user_api="blas")
    except:
        # threadpoolctl not available, need a dummy context manager
        import contextlib
        cm = contextlib.nullcontext()
    with cm:
        hit = div[0,0] > lim0
        # Get the condition number of each pixel
        work    = np.ascontiguousarray(div[:,:,hit].T)
        E, V    = np.linalg.eigh(work)
        cond    = E[:,0]/E[:,-1]
        good    = cond >= lim
        # Invert the good ones
        inv_good= np.einsum("...ij,...j,...kj->...ik", V[good], 1/E[good], V[good])
        # Treat the bad ones as being purely T
        inv_bad = work[~good]*0
        inv_bad[:,0,0] = 1/work[~good,0,0]
        # Copy back
        work[good]  = inv_good
        work[~good] = inv_bad
        # And put into final output
        idiv = div*0
        idiv[:,:,hit] = work.T
    return idiv


def measure_cov(d, nmax=10000):
    d = d[:,::max(1,d.shape[1]//nmax)]
    n,m = d.shape
    step  = 10000
    res = np.zeros((n,n),d.dtype)
    for i in range(0,m,step):
        sub = mycontiguous(d[:,i:i+step])
        res += np.real(sub.dot(np.conj(sub.T)))
    return res/m

def project_out(d, modes): return d-modes.T.dot(modes.dot(d))

def project_out_from_matrix(A, V):
    # Used Woodbury to project out the given vectors from the covmat A
    if V.size == 0: return A
    Q = A.dot(V)
    return A - Q.dot(np.linalg.solve(np.conj(V.T).dot(Q), np.conj(Q.T)))

def measure_power(d): return np.real(np.mean(d*np.conj(d),-1))

def makebins(edge_freqs, srate, nfreq, nmin=0, rfun=None, cap=True):
    # Urk, this function is ugly. It's an old one from when I
    # first started learning python.
    # Translate from frequency to index
    binds = freq2ind(edge_freqs, srate, nfreq, rfun=rfun)
    if cap: binds = np.concatenate([[0],binds,[nfreq]])
    # Cap at nfreq and eliminate any resulting empty bins
    binds = np.unique(np.minimum(binds,nfreq))
    # Make sure no bins have two few entries
    if nmin > 0:
        binds2 = [binds[0]]
        for b in binds[1:-1]:
            if b-binds2[-1] >= nmin: binds2.append(b)
        binds2.append(binds[-1])
        # If the last bin is too short, remove the second-to-last entry
        if binds2[-1]-binds2[-2] < nmin:
            del binds2[-2]
        binds = binds2
    # Go from edges to [:,{from,to}]
    bins  = np.array([binds[:-1],binds[1:]]).T
    return bins

def mycontiguous(a):
    # I used this in act for some reason, but not sure why. I vaguely remember ascontiguousarray
    # causing weird failures later in lapack
    b = np.zeros(a.shape, a.dtype)
    b[...] = a[...]
    return b

def find_modes_jon(ft, bins, eig_lim=None, single_lim=0, skip_mean=False, verbose=False):
    ndet = ft.shape[0]
    vecs = np.zeros([ndet,0])
    if not skip_mean:
        # Force the uniform common mode to be included. This
        # assumes all the detectors have accurately measured gain.
        # Forcing this avoids the possibility that we don't find
        # any modes at all.
        vecs = np.concatenate([vecs,np.full([ndet,1],ndet**-0.5)],1)
    for bi, b in enumerate(bins):
        d    = ft[:,b[0]:b[1]]
        cov  = measure_cov(d)
        cov  = project_out_from_matrix(cov, vecs)
        e, v = np.linalg.eig(cov)
        e, v = e.real, v.real
        #e, v = e[::-1], v[:,::-1]
        accept = np.full(len(e), True, bool)
        if eig_lim is not None:
            # Compute median, exempting modes we don't have enough data to measure
            nsamp    = b[1]-b[0]+1
            median_e = np.median(np.sort(e)[::-1][:nsamp])
            accept  &= e/median_e >= eig_lim[bi]
        if verbose: print("bin %d: %4d modes above eig_lim" % (bi, np.sum(accept)))
        if single_lim is not None and e.size:
            # Reject modes too concentrated into a single mode. Since v is normalized,
            # values close to 1 in a single component must mean that all other components are small
            singleness = np.max(np.abs(v),0)
            accept    &= singleness < single_lim[bi]
        if verbose: print("bin %d: %4d modes also above single_lim" % (bi, np.sum(accept)))
        e, v = e[accept], v[:,accept]
        vecs = np.concatenate([vecs,v],1)
    return vecs

def measure_detvecs(ft, vecs, nper=2):
    # Allow too narrow bins
    nfull = vecs.shape[1]
    n     = min(nfull, ft.shape[-1]//nper+1)
    vecs  = vecs[:,:n]
    # Measure amps when we have non-orthogonal vecs
    rhs  = vecs.T.dot(ft)
    div  = vecs.T.dot(vecs)
    amps = np.linalg.solve(div,rhs)
    E    = np.mean(np.abs(amps)**2,1)
    # Project out modes for every frequency individually
    dclean = ft - vecs.dot(amps)
    # The rest is assumed to be uncorrelated
    Nu = np.mean(np.abs(dclean)**2,1)
    # The total auto-power
    Nd = np.mean(np.abs(ft)**2,1)
    # Expand E to full requested len with low but
    # non-zero power so we don't need to special-case elsewhere
    small = np.min(Nu)*1e-6
    E  = np.pad(E, (0,nfull-n), constant_values=small)
    return E, Nu, Nd

def sichol(A):
    iA = np.linalg.inv(A)
    try: return np.linalg.cholesky(iA), 1
    except np.linalg.LinAlgError:
        return np.linalg.cholesky(-iA), -1

def safe_inv(a):
    with putils.nowarn():
        res = 1/a
        res[~np.isfinite(res)] = 0
    return res

def woodbury_invert(D, V, s=1):
    """Given a compressed representation C = D + sVV', compute a
    corresponding representation for inv(C) using the Woodbury
    formula."""
    V, D = map(np.asarray, [V,D])
    ishape = D.shape[:-1]
    # Flatten everything so we can be dimensionality-agnostic
    D = D.reshape(-1, D.shape[-1])
    V = V.reshape(-1, V.shape[-2], V.shape[-1])
    I = np.eye(V.shape[2])
    # Allocate our output arrays
    iD = safe_inv(D)
    iV = V*0
    # Invert each
    for i in range(len(D)):
        core = I*s + (V[i].T*iD[i,None,:]).dot(V[i])
        core, sout = sichol(core)
        iV[i] = iD[i,:,None]*V[i].dot(core)
    sout = -sout
    return iD, iV, sout

def apply_window(tod, nsamp, exp=1):
    """Apply a cosine taper to each end of the TOD."""
    if nsamp <= 0: return
    taper   = 0.5*(1-np.cos(np.arange(1,nsamp+1)*np.pi/nsamp))
    taper **= exp
    tod[...,:nsamp]  *= taper
    tod[...,-nsamp:] *= taper[::-1]

########################
###### Utilities #######
########################

def get_ids(query, context=None):
    try:
        with open(query, "r") as fname:
            return [line.split()[0] for line in fname]
    except FileNotFoundError:
        return context.obsdb.query(query or "1")['obs_id']

def get_subids(query, context=None, method="auto"):
    """A subid has the form obs_id:wafer_slot:band, and is a natural
    unit to use for mapmaking."""
    if method == "auto":
        try: return get_subids_file(query, context=context)
        except FileNotFoundError: return get_subids_query(query, context=context)
    elif method == "file":
        return get_subids_file(query, context=context)
    elif method == "query":
        return get_subids_query(query, context=context)
    else:
        raise ValueError("Unrecognized method for get_subids: '%s'" % (str(method)))

def get_subids_query(query, context):
    obs_ids = context.obsdb.query(query or "1")['obs_id']
    sub_ids = expand_ids(obs_ids, context)
    return sub_ids

def get_subids_file(fname, context=None):
    with open(fname, "r") as fname:
        sub_ids = [line.split()[0] for line in fname]
    sub_ids = expand_ids(sub_ids, context=context)
    return sub_ids

def expand_ids(obs_ids, context=None, bands=None):
    """Given a list of ids that are either obs_ids or sub_ids, expand any obs_ids
    into sub_ids and return the resulting list.

    To infer the bands available either bands or context must be passed.
    If bands is passed then it should be a list of the multichroic bands
    available for all the wafers. Example: bands=["f090","f150"].

    Otherwise a Context should be passed, and the bands will be inferred
    per obs by querying its obsdb. This is the standard case.
    """
    if len(obs_ids) == 0: return []
    # Get the tube flavor for each. We will need this to get the bands.
    if context is not None:
        # Make sure we have plain obs_id regardless of whether obs_ids or sub_ids were passed
        actual_obs_ids = np.char.partition(obs_ids, ":")[:,0]
        # Get the tube flavor for each, and their meanings
        all_ids, flavors = [], []
        for row in context.obsdb.conn.execute("select obs_id, tube_flavor from obs"):
            all_ids.append(row[0])
            flavors.append(row[1])
        all_ids = np.array(all_ids)
        inds    = putils.find(all_ids, actual_obs_ids)
        flavors = [flavors[ind].lower() for ind in inds]
        flavor_map = {
            "lf": ("f030", "f040"),
            "mf": ("f090", "f150"),
            "hf": ("f150", "f220"),
            "uhf": ("f220", "f280"),
            None: ("f000",),
        }
    elif bands is not None:
        flavors    = ["a"]*len(obs_ids)
        flavor_map = {"a": bands}
    else:
        raise ValueError("Either bands or context must be passed")
    # Then loop through each and expand as necessary
    sub_ids = []
    for obs_id, flavor in zip(obs_ids, flavors):
        bands  = flavor_map[flavor]
        toks   = obs_id.split(":")
        if len(toks) == 3:
            # Already sub_id
            sub_ids.append(obs_id)
        elif len(toks) != 1:
            raise ValueError("Invalid obs_id '%s'" % (str(obs_id)))
        else:
            toks = obs_id.split("_")
            # SAT format: obs_1719902396_satp1_1111111
            # LAT format: 1696118940_i3_100 (why no obs in front?)
            wafer_mask = toks[-1]
            # Loop through wafer slots
            for si, status in enumerate(wafer_mask):
                if status == "1":
                    for band in bands:
                        sub_ids.append("%s:ws%d:%s" % (obs_id, si, band))
    return sub_ids

def filter_subids(subids, wafers=None, bands=None):
    subids = np.asarray(subids)
    if wafers is not None:
        wafs   = astr_tok(subids,":",1)
        subids = subids[np.isin(wafs, wafers)]
    if bands is not None:
        bpass  = astr_tok(subids,":",2)
        subids = subids[np.isin(bpass, bands)]
    return subids

def astr_cat(*arrs):
    res = np.char.add(arrs[0], arrs[1])
    for arr in arrs[2:]:
        res = np.char.add(res,arr)
    return res

def astr_tok(astr, sep, i):
    for j in range(i):
        astr = np.char.partition(astr,sep)[:,2]
    return np.char.partition(astr,sep)[:,0]

def split_subids(subids):
    ids, _, rest   = np.char.partition(subids, ":").T
    wafs, _, bands = np.char.partition(rest, ":").T
    return ids, wafs, bands

def infer_comps(ncomp): return ["T","QU","TQU"][ncomp-1]

def parse_recentering(desc):
    """Parse an object centering description, as provided by the --center-at argument.
    The format is [from=](ra:dec|name),[to=(ra:dec|name)],[up=(ra:dec|name|system)]
    from: specifies which point is to be centered. Given as either
      * a ra:dec pair in degrees
      * the name of a pre-defined celestial object (e.g. Saturn), which should not move
        appreciably in celestial coordinates during a TOD
    to: the point at which to recenter. Optional. Given as either
      * a ra:dec pair in degrees
      * the name of a pre-defined celestial object
      Defaults to ra=0,dec=0 or ra=0,dec=90, depending on the projection
    up: which direction should point up after recentering. Optional. Given as either
      * the name of a coordinate system (e.g. hor, cel, gal), in which case
        up will point towards the north pole of that system
      * a ra:dec pair in degrees
      * the name of a pre-defined celestial object
      Defualts to the celestial north pole

    Returns "info", a bunch representing the recentering specification in more python-friendly
    terms. This can later be passed to evaluate_recentering to get the actual euler angles that perform
    the recentering.

    Examples:
      * 120.2:-13.8
        Centers on ra = 120.2°, dec = -13.8°, with up being celestial north
      * Saturn
        Centers on Saturn, with up being celestial north
      * Uranus,up=hor
        Centers on Uranus, but up is up in horizontal coordinates. Appropriate for beam mapping
      * Uranus,up=hor,to=0:90
        As above, but explicitly recenters on the north pole
    """
    # If necessary the syntax above could be extended with from_sys, to_sys and up-sys, which
    # so one could specify galactic coordiantes for example. Or one could generalize
    # from ra:dec to phi:theta[:sys], where sys would default to cel. But for how I think
    # this is enough.
    args = desc.split(",")
    info  = {"to":"auto", "up":"cel", "from_sys":"cel", "to_sys":"cel", "up_sys":"cel"}
    for ai, arg in enumerate(args):
        # Split into key,value
        toks = arg.split("=")
        if ai == 0 and len(toks) == 1:
            key, val = "from", toks[0]
        elif len(toks) == 2:
            key, val = toks
        else:
            raise ValueError("parse_recentering wants key=value format, but got %s" % (arg))
        # Handle the values
        if ":" in val:
            val = [float(w)*putils.degree for w in val.split(":")]
        info[key] = val
    if "from" not in info:
        raise ValueError("parse_recentering needs at least the from argument")
    return info

def evaluate_recentering(info, ctime, geom=None, site=None, weather="typical"):
    """Evaluate the quaternion that performs the coordinate recentering specified in
    info, which can be obtained from parse_recentering."""
    import ephem

    # Get the coordinates of the from, to and up points. This was a bit involved...
    def to_cel(lonlat, sys, ctime=None, site=None, weather=None):
        # Convert lonlat from sys to celestial coorinates. Maybe polish and put elswhere
        if sys == "cel" or sys == "equ":
            return lonlat
        elif sys == "hor":
            return so3g.proj.CelestialSightLine.az_el(ctime, lonlat[0], lonlat[1], site=site, weather=weather).coords()[0,:2]
        else:
            raise NotImplementedError
    def get_pos(name, ctime, sys=None):
        if isinstance(name, str):
            if name in ["hor", "cel", "equ", "gal"]:
                return to_cel([0,np.pi/2], name, ctime, site, weather)
            elif name == "auto":
                return np.array([0,0]) # would use geom here
            else:
                obj = getattr(ephem, name)()
                djd = ctime/86400 + 40587.0 + 2400000.5 - 2415020
                obj.compute(djd)
                return np.array([obj.a_ra, obj.a_dec])
        else:
            return to_cel(name, sys, ctime, site, weather)
    p1 = get_pos(info["from"], ctime, info["from_sys"])
    p2 = get_pos(info["to"],   ctime, info["to_sys"])
    pu = get_pos(info["up"],   ctime, info["up_sys"])
    return [p1,p2,pu]

def recentering_to_quat_lonlat(p1, p2, pu):
    """Return the quaternion that represents the rotation that takes point p1
    to p2, with the up direction pointing towards the point pu, all given as lonlat pairs"""
    from so3g.proj import quat

    # 1. First rotate our point to the north pole: Ry(-(90-dec1))Rz(-ra1)
    # 2. Apply the same rotation to the up point.
    # 3. We want the up point to be upwards, so rotate it to ra = 180°: Rz(pi-rau2)
    # 4. Apply the same rotation to the real point
    # 5. Rotate the point to its target position: Rz(ra2)Ry(90-dec2)
    ra1, dec1 = p1
    ra2, dec2 = p2
    rau, decu = pu
    qu    = quat.rotation_lonlat(rau, decu)
    R     = ~quat.rotation_lonlat(ra1, dec1)
    rau2  = quat.decompose_lonlat(R*qu)[0]
    R     = quat.euler(2, ra2)*quat.euler(1, np.pi/2-dec2)*quat.euler(2, np.pi-rau2)*R
    a = quat.decompose_lonlat(R*quat.rotation_lonlat(ra1,dec1))
    return R

def highpass(tod, fknee=1e-2, alpha=3):
    ft   = fft.rfft(tod)
    freq = fft.rfftfreq(tod.shape[1])
    ft  /= 1 + (freq/fknee)**-alpha
    return fft.irfft(ft, tod, normalize=True)

def find_boresight_jumps(vals, width=20, tol=0.1):
    # median filter array to get reference behavior
    bad   = np.zeros(vals.size,dtype=bool)
    width = int(width)//2*2+1
    fvals = putils.block_mean_filter(vals, width)
    bad  |= np.abs(vals-fvals) > tol
    return bad

def robust_unwind(a, period=2*np.pi, cut=None, tol=1e-3, mask=None):
    """Like putils.unwind, but only registers something as an angle jump if
    it is of just the right shape. If cut is specified, it should be a list
    of valid angle cut positions, which will further restrict when jumps are
    allowed. Only 1d input is supported."""
    # Find places where a jump would be acceptable
    period = float(period)
    diffs  = (a[1:]-a[:-1])/period
    valid  = np.abs(np.abs(diffs)-1) < tol/period
    diffs *= valid
    jumps  = np.concatenate([[0],np.round(diffs)])
    if mask is not None:
        jumps[mask] = 0
        jumps[:-1][mask[1:]] = 0
    if cut is not None:
        near_cut = np.zeros(a.size, bool)
        for cutval in cut:
            near_cut |= np.abs((a - cutval + period/2) % period + period/2) < tol
        jumps[~near_cut] = 0
    # Then correct our values
    return a - np.cumsum(jumps)*period

def find_elevation_outliers(el, tol=0.5*putils.degree):
    typ = np.median(el[::100])
    return np.abs(el-typ)>tol

def freq2ind(freqs, srate, nfreq, rfun=None):
    """Returns the index of the first fourier mode with greater than freq
    frequency, for each freq in freqs."""
    if freqs is None: return freqs
    if rfun  is None: rfun = np.ceil
    return rfun(np.asarray(freqs)/(srate/2.0)*nfreq).astype(int)

def rangemat_sum(rangemat):
    res = np.zeros(len(rangemat))
    for i, r in enumerate(rangemat):
        ra = r.ranges()
        res[i] = np.sum(ra[:,1]-ra[:,0])
    return res

def find_usable_detectors(obs, maxcut=0.1, glitch_flags: str = "flags.glitch_flags"):
    ncut  = rangemat_sum(obs[glitch_flags])
    good  = ncut < obs.samps.count * maxcut
    return obs.dets.vals[good]

def fix_boresight_glitches(obs, ang_tol=0.1*putils.degree, t_tol=1):
    az   = robust_unwind(obs.boresight.az)
    bad  = find_boresight_jumps(az,               tol=ang_tol)
    bad |= find_boresight_jumps(obs.boresight.el, tol=ang_tol)
    bad |= find_boresight_jumps(obs.timestamps,   tol=t_tol)
    bcut = so3g.RangesInt32.from_mask(bad)
    obs.boresight.az[:] = az
    tod_ops.get_gap_fill_single(obs.timestamps,   bcut, swap=True)
    tod_ops.get_gap_fill_single(obs.boresight.az, bcut, swap=True)
    tod_ops.get_gap_fill_single(obs.boresight.el, bcut, swap=True)

def unarr(a): return np.array(a).reshape(-1)[0]

def downsample_ranges(ranges, down):
    """Downsample either an array of ranges [:,{from,to}]
    or an so3gRangesInt32 object, by the integer factor down.
    The downsampling is inclusive: The output ranges will be
    as small as possible while fully encompassing the input ranges."""
    if isinstance(ranges, so3g.RangesInt32):
        return so3g.RangesInt32.from_array(downsample_ranges(ranges.ranges(), down), (ranges.count+down-1)//down)
    oranges = ranges.copy()
    # Lower range is simple
    oranges[:,0] //= down
    # End should be one above highest impacted index.
    oranges[:,1] = (oranges[:,1]-1)//down+1
    return oranges

def downsample_cut(cut, down):
    """Given an integer RangesMatrix respresenting samples to cut,
    return a new such RangesMatrix that describes which samples to
    cut if the timestream were to be downsampled by the integer
    factor down."""
    return so3g.proj.ranges.RangesMatrix([downsample_ranges(r,down) for r in cut.ranges])

def downsample_obs(obs, down, skip_signal=False):
    """Downsample AxisManager obs by the integer factor down.

    This implementation is quite specific and probably needs
    generalization in the future, but it should work correctly
    and efficiently for ACT-like data at least. In particular
    it uses fourier-resampling when downsampling the detector
    timestreams to avoid both aliasing noise and introducing
    a transfer function."""
    assert down == putils.nint(down), "Only integer downsampling supported, but got '%.8g'" % down
    if down == 1: return obs
    # Compute how many samples we will end up with
    onsamp = (obs.samps.count+down-1)//down
    # Set up our output axis manager
    axes   = [obs[axname] for axname in obs._axes if axname != "samps"]
    res    = core.AxisManager(core.IndexAxis("samps", onsamp), *axes)
    # Stuff without sample axes
    for key, axes in obs._assignments.items():
        if "samps" not in axes:
            res.wrap(*get_wrappable(obs, key))
    # The normal sample stuff
    res.wrap("timestamps", obs.timestamps[::down], [(0, "samps")])
    bore = core.AxisManager(core.IndexAxis("samps", onsamp))
    for key in ["az", "el", "roll"]:
        bore.wrap(key, getattr(obs.boresight, key)[::down], [(0, "samps")])
    res.wrap("boresight", bore)
    if not skip_signal:
        res.wrap("signal", resample.resample_fft_simple(obs.signal, onsamp), [(0,"dets"),(1,"samps")])

    # TODO: Once it's finalized whether cuts are in obs.flags.X or obs.X, remove these
    # if tests
    cut_keys = []
    if "glitch_flags"  in obs:
        cut_keys.append("glitch_flags")
    elif "flags.glitch_flags" in obs:
        cut_keys.append("flags.glitch_flags")

    if "source_flags" in obs:
        cut_keys.append("source_flags")
    elif "flags.source_flags" in obs:
        cut_keys.append("flags.source_flags")

    # We need to add a res.flags FlagManager to res
    res = res.wrap('flags', core.FlagManager.for_tod(res))
    for key in cut_keys:
        new_key = key.split(".")[-1]
        res.flags.wrap(new_key, downsample_cut(obs[key], down),
                       [(0,"dets"),(1,"samps")])

    # Not sure how to deal with flags. Some sort of or-binning operation? But it
    # doesn't matter anyway
    return res

def get_wrappable(axman, key):
    val = getattr(axman, key)
    if isinstance(val, core.AxisManager):
        return key,val
    else:
        axes   = axman._assignments[key]
        axdesc = [(k,v) for k,v in enumerate(axes) if v is not None]
        return key, val, axdesc

def get_flags(obs, flagnames):
    """Parse detector-set splits"""
    cuts_out = None
    if flagnames is None:
        return so3g.proj.RangesMatrix.zeros(obs.shape)
    det_splits = ['det_left','det_right','det_in','det_out','det_upper','det_lower']
    for flagname in flagnames:
        if flagname in det_splits:
            cuts = obs.det_flags[flagname]
        elif flagname == 'scan_left':
            cuts = obs.flags.left_scan
        elif flagname == 'scan_right':
            cuts = obs.flags.right_scan
        else:
            cuts = getattr(obs.flags, flagname) # obs.flags.flagname

        ## Add to the output matrix
        if cuts_out is None:
            cuts_out = cuts
        else:
            cuts_out += cuts
    return cuts_out

def import_optional(module_name):
    try:
        module = importlib.import_module(module_name)
        return module
    except:
        return None

def setup_passes(downsample="1", maxiter="500", interpol="nearest", npass=None):
    """Set up information multipass mapmaking. Supports arguments
    of the form num or num,num,.... Infers the number of passes
    from the maximum length of these (or npass if explicitly given),
    and makes sure they all have this length by duplicating the item
    as necessary.

    Example:
     setup_passes(downsample="4,4,1", maxiter="300,300,50", interpol="linear")
     returns bunch.Bunch(downsample=[4,4,1], maxiter=[300,300,50],
      interpol=["linear","linear","linear"])"""
    tmp            = bunch.Bunch()
    tmp.downsample = putils.parse_ints(downsample)
    tmp.maxiter    = putils.parse_ints(maxiter)
    tmp.interpol   = interpol.split(",")
    # The entries may have different lengths. We use the max
    # and then pad the others by repeating the last element.
    # The final output will be a list of bunches
    if npass is None:
        npass = max([len(tmp[key]) for key in tmp])
    passes    = []
    for i in range(npass):
        entry = bunch.Bunch()
        for key in tmp:
            entry[key] = tmp[key][min(i,len(tmp[key])-1)]
        passes.append(entry)
    return passes

Base = declarative_base()

class AtomicInfo(Base):
    __tablename__ = "atomic"

    obs_id: Mapped[str] = mapped_column(primary_key=True)
    telescope: Mapped[str] = mapped_column(primary_key=True)
    freq_channel: Mapped[str] = mapped_column(primary_key=True)
    wafer: Mapped[str] = mapped_column(primary_key=True)
    ctime: Mapped[int] = mapped_column(primary_key=True)
    split_label: Mapped[str] = mapped_column(primary_key=True)
    valid: Mapped[Optional[bool]]
    split_detail: Mapped[Optional[str]]
    prefix_path: Mapped[Optional[str]]
    elevation: Mapped[Optional[float]]
    azimuth: Mapped[Optional[float]]
    pwv: Mapped[Optional[float]]
    dpwv: Mapped[Optional[float]]
    total_weight_qu: Mapped[Optional[float]]
    mean_weight_qu: Mapped[Optional[float]]
    median_weight_qu: Mapped[Optional[float]]
    leakage_avg: Mapped[Optional[float]]
    noise_avg: Mapped[Optional[float]]
    ampl_2f_avg: Mapped[Optional[float]]
    gain_avg: Mapped[Optional[float]]
    tau_avg: Mapped[Optional[float]]
    f_hwp: Mapped[Optional[float]]
    roll_angle: Mapped[Optional[float]]
    scan_speed: Mapped[Optional[float]]
    scan_acc: Mapped[Optional[float]]
    sun_distance: Mapped[Optional[float]]
    ambient_temperature: Mapped[Optional[float]]
    uv: Mapped[Optional[float]]
    ra_center: Mapped[Optional[float]]
    dec_center: Mapped[Optional[float]]
    number_dets: Mapped[Optional[int]]
    moon_distance: Mapped[Optional[float]]
    wind_speed: Mapped[Optional[float]]
    wind_direction: Mapped[Optional[float]]

    def __init__(self, obs_id, telescope, freq_channel, wafer, ctime, split_label):
        self.obs_id = obs_id
        self.telescope = telescope
        self.freq_channel = freq_channel
        self.wafer = wafer
        self.ctime = ctime
        self.split_label = split_label

    def __repr__(self):
        return f"({self.obs_id},{self.telescope},{self.freq_channel},{self.wafer},{self.ctime},{self.split_label})"

    @classmethod
    def from_dict(cls, contents: dict):
        core = cls(
            obs_id=contents.pop("obs_id"),
            telescope=contents.pop("telescope"),
            freq_channel=contents.pop("freq_channel"),
            wafer=contents.pop("wafer"),
            ctime=contents.pop("ctime"),
            split_label=contents.pop("split_label"),
        )
        for k, v in contents.items():
            setattr(core, k, v)
        return core

def atomic_db_aux(atomic_db, info: list[AtomicInfo]):
    engine = create_engine("sqlite:///%s" % atomic_db, echo=False)
    Base.metadata.create_all(bind=engine)
    Session = sessionmaker(bind=engine)
    with Session() as session:
        session.add_all(info)
        try:
            session.commit()
        except exc.IntegrityError:
            session.rollback()



from sotodlib import core

import collections
import fnmatch
import logging
import os
import numpy as np
import warnings

from . import ResultSet

logger = logging.getLogger(__name__)

REGISTRY = {
    '_default': 'DefaultHdf'
}


class LoaderError(RuntimeError):
    """
    Use with two args: (pithy_summary, formatted_detail)
    """


class IncompleteMetadataError(RuntimeError):
    pass


class IncompleteDetInfoError(RuntimeError):
    pass


class MetadataSpecError(RuntimeError):
    pass


class MetadataSpecWarning(Warning):
    pass


class SuperLoader:
    def __init__(self, context=None, detdb=None, obsdb=None, working_dir=None):
        """Metadata batch loader.

        Args:
          context (Context): context, from which detdb and obsdb will
            be pulled unless they are specified explicitly.
          detdb (DetDb): detdb to use when resolving detector axis.
          obsdb (ObsDb): obsdb to use when resolving obs axis.
          working_dir (str): base directory for any metadata specified
            as relative paths.  If None, but if context is not None,
            then the path of context.filename is used; otherwise cwd
            is used.

        """
        if context is not None:
            if detdb is None:
                detdb = context.detdb
            if obsdb is None:
                obsdb = context.obsdb
            if working_dir is None and context.filename is not None:
                working_dir = os.path.split(context.filename)[0]
        if working_dir is None:
            working_dir = os.getcwd()
        self.detdb = detdb
        self.obsdb = obsdb
        self.manifest_cache = {}
        self.working_dir = working_dir

    @staticmethod
    def register_metadata(name, loader_class):
        """Globally register a metadata "Loader Class".

        Args:
          name (str): Name under which to register the loader.
            Metadata archives will request the loader class using
            this name.
          loader_class: Metadata loader class.

        """
        REGISTRY[name] = loader_class

    def load_one(self, spec, request, det_info):
        """Process a single metadata entry (spec) by loading a ManifestDb and
        reading metadata for a particular observation.  The request
        must be pre-augmented with all ObsDb info that might be
        needed.  det_info is used to screen the returned data for the
        various index_lines.

        Args:
          spec (dict): A metadata specification dict (corresponding to
            a metadata list entry in context.yaml), or MetadataSpec
            object.
          request (dict): A metadata request dict (stating what
            observation and detectors are of interest).
          det_info (ResultSet): Table of detector properties to use
            when resolving metadata that is indexed with dets:...
            fields.

        Notes:
          If passing ``spec`` as a dict, see the schema described in
          :class:`MetadataSpec`.

          Any filenames in the ManifestDb that are given as relative
          paths will be resolved relative to the directory where the
          db file is found.

          The ``request`` dict specifies what times and detectors are
          of interest.  If the metadata archive is indexed by
          timestamp and wafer_slot, then you might pass in::

             {'obs:timestamp': 1234567000.,
              'dets:wafer_slot': 'w01'}

          When this function is invoked from self.load, the request
          dict will have been automatically "augmented" using the
          ObsDb.  The main purpose of this is to provide obs:timestamp
          (and any other useful indexing fields) from ObsDb based on
          obs:obs_id.

          The ``det_info`` object comes into play in cases where a
          loaded metadata result refers to some large group of
          detectors, but the metadata index, or the user request,
          expresses that the result should be limited to only a subset
          of those detectors.  This is notated in practice by
          including dets:... fields in the index data in the ManifestDb,
          or in the request dict.  Only fields already present in
          det_info may be included in the request dict.

        Returns:
          A list of tuples (unpacker, item), corresponding to each
          entry in spec_list.  The unpacker is an Unpacker object
          created based on the 'name' field.  The item is the metadata
          in its native format (which could be a ResultSet or
          AxisManager), with all restrictions specified in request
          already applied.

        """
        if isinstance(spec, dict):
            spec = MetadataSpec.from_dict(spec)

        # Load the database, match the request,
        if isinstance(spec.db, str):
            # The usual case.
            dbfile = os.path.join(self.working_dir, spec.db)
            dbpath = os.path.split(dbfile)[0]
            if dbfile not in self.manifest_cache:
                if dbfile.endswith('sqlite'):
                    man = core.metadata.ManifestDb.readonly(dbfile)
                else:
                    man = core.metadata.ManifestDb.from_file(dbfile)
                self.manifest_cache[dbfile] = man
            man = self.manifest_cache[dbfile]
        elif isinstance(spec.db, core.metadata.ManifestDb):
            # Useful for testing and hacking
            dbpath = self.working_dir
            man = spec.db

        # Do we have all the keys we need?
        required_obs_keys = _filter_items(
            'obs:', man.scheme.get_required_params(), remove=False)
        missing_obs_keys = (set(required_obs_keys) - set(request.keys()))
        if len(missing_obs_keys):
            raise RuntimeError(
                f'Metadata request is indexed by {request.keys()} but '
                f'ManifestDb requires {required_obs_keys}.')

        required_dets_keys = _filter_items(
            'dets:', man.scheme.get_required_params(), remove=False)
        missing_dets_keys = list((set(required_dets_keys) - set(request.keys())))

        if len(missing_dets_keys):
            # Make request to ManifestDb for each detector bundle.
            short_keys = _filter_items('dets:', missing_dets_keys)
            try:
                subreqs = det_info.subset(keys=short_keys).distinct()
            except:
                raise RuntimeError(
                    f'Metadata request requires keys={missing_dets_keys} '
                    f'but det_info={det_info}.')
            subreqs.keys = missing_dets_keys # back with dets: prefix ...
        else:
            subreqs = ResultSet([], [()])  # length 1!

        index_lines = []
        for subreq in subreqs:
            # Reject any subreq that explicitly contradicts request on any key.
            if any([subreq.get(k, v) != v for k, v in request.items()]):
                continue
            subreq.update(request)

            try:
                _lines = man.match(subreq, multi=True, prefix=dbpath)
            except Exception as e:
                text = str(e)
                raise LoaderError('Exception when matching subrequest.',
                                  f"An exception occurred while processing sub-request:\n\n"
                                  f"  subreq={subreq}\n\n")
            for _line in _lines:
                # Now reject any _line if they contradict subreq.
                if any([subreq.get(k, v) != v for k, v in _line.items()]):
                    continue
                _line.update(subreq)
                index_lines.append(_line)

        if len(index_lines) == 0:
            raise LoaderError(
                f'There were no metadata records found for this observation ({request}).')

        # Pre-screen the index_lines for dets:* assignments; plan to
        # skip lines that aren't relevant according to det_info.
        to_skip = []
        for index_line in index_lines:
            logger.debug(f'Pre-screening index_line={index_line}')
            skip_this = len(det_info) == 0
            if not skip_this:
                mask = np.ones(len(det_info), bool)
                for k, v in _filter_items('dets:', index_line, remove=True).items():
                    if k not in det_info.keys:
                        raise IncompleteDetInfoError(
                            f"Entry requires det_info['{k}'], and that is not defined.")
                    mask *= (det_info[k] == v)
                skip_this = (mask.sum() == 0)
            to_skip.append(skip_this)

        if len(index_lines) == 0:
            # If we come out the other side with no data to load,
            # invent one so that we at least get the structure of the
            # metadata (even though we'll throw out all the actual
            # results).  You can get here if someone passes dets=[].
            candidate_index_lines = man.inspect(request, False, prefix=dbpath)
            index_lines.append(candidate_index_lines[0])
            to_skip = [False]

        elif all(to_skip):
            # Load at least one item, or we won't have the structure of
            # the output.
            to_skip[0] = False

        # Load each index_line.
        results = []
        for skip, index_line in zip(to_skip, index_lines):
            if skip:
                logger.debug(f'Skipping load for index_line={index_line}')
                continue
            logger.debug(f'Loading for index_line={index_line}')

            loader = spec.loader
            if loader is None:
                loader = index_line.get('loader', REGISTRY['_default'])
            try:
                loader_class = REGISTRY[loader]
            except KeyError:
                raise LoaderError(
                    'Loader function not found.',
                    f'No metadata loader registered under name "{loader}"')

            loader_object = loader_class()  # pass obs info?
            loader_kwargs = {}
            if spec.load_fields is not None:
                loader_kwargs['load_fields'] = spec.load_fields
            mi1 = loader_object.from_loadspec(index_line, **loader_kwargs)

            # Restrict returned values according to the specs in index_line.

            if isinstance(mi1, ResultSet):
                # For simple tables, the restrictions can be
                # integrated into the table, to be dealt with later.
                det_restricts = _filter_items('dets:', index_line, remove=False)
                mask = np.ones(len(mi1), bool)
                keep_cols = list(mi1.keys)
                new_cols = []
                for k, v in det_restricts.items():
                    if k in mi1.keys:
                        mask *= (mi1[k] == v)
                    else:
                        new_cols.append((k, v))
                a = mi1.subset(keys=keep_cols, rows=mask)
                mi2 = ResultSet([k for k, v in new_cols],
                                [[v for k, v in new_cols]] * len(a))
                mi2.merge(a)

            elif isinstance(mi1, core.AxisManager):
                # For AxisManager, allow user to drop some items
                # before proceeding.  This only catches fields at root
                # level of the AxisManager.
                for pattern in spec.drop_fields:
                    to_drop = fnmatch.filter(mi1._fields.keys(), pattern)
                    for k in to_drop:
                        mi1.move(k, None)

                # If the dets axis is present, it *must* reconcile
                # 1-to-1 with some field in det_info, and that may be
                # used to toss things out based on index_line.
                if 'dets' in mi1:
                    det_restricts = _filter_items('dets:', index_line, remove=True)
                    dets_key = 'readout_id'
                    new_dets, i_new, i_info = core.util.get_coindices(
                        mi1.dets.vals, det_info[dets_key])

                    mask = np.ones(len(i_new), bool)
                    if len(i_info):
                        for k, v in det_restricts.items():
                            mask *= (det_info[k][i_info] == v)
                    if mask.all() and len(new_dets) == mi1.dets.count:
                        mi2 = mi1
                    else:
                        mi2 = mi1.restrict('dets', new_dets[mask])
                else:
                    mi2 = mi1

            else:
                raise LoaderError(
                    'Invalid metadata carrier.',
                    'Returned object is non-specialized type {}: {}'
                    .format(mi1.__class__, mi1))

            results.append(mi2)

        # Check that we got results, then combine them in to single ResultSet.
        logger.debug(f'Concatenating {len(results)} results: {results}')
        assert(len(results) > 0)
        if len(results) == 1:
            result = results[0]
        else:
            result = results[0].concatenate(results)
        return result

    def load(self, spec_list, request, det_info=None, free_tags=[],
             free_tag_fields=[], dest=None, check=False, det_info_scan=False,
             ignore_missing=False, on_missing=None):
        """Loads metadata objects and processes them into a single
        AxisManager.

        Args:
          spec_list (list of dicts): Each dict is a metadata spec, as
            described in load_one.
          request (dict): A request dict.
          det_info (AxisManager): Detector info table to use for
            reconciling 'dets:...' field restrictions.
          free_tags (list of str): Strings that restrict the detector
            to any detector that matches the string in any of the
            det_info fields listed in free_tag_fields.
          free_tag_fields (list of str): Fields (of the form dets:x)
            that can be inspected to match free_tags.
          dest (AxisManager or None): Destination container for the
            metadata (if None, a new one is created).
          check (bool): If True, run in check mode (see Notes).
          det_info_scan (bool): If True, *only* process entries that
            directly update det_info.
          ignore_missing (bool): If True, don't fail when a metadata
            item can't be loaded, just try to proceed without it.
          on_missing (dict): If a key here matches the label of a
            metadata entry, the value will override the on_missing
            entry of the metadata entry.  (Each value must be "trim",
            "skip" or "fail".)

        Returns:
          In normal mode, an AxisManager containing the metadata
          (dest).  In check mode, a list of tuples (spec, exception).

        Notes:
          If check=True, this won't store and return the loaded
          metadata; it will instead return a list of the same length
          as spec_list, with either None (if the entry loaded
          successful) or the Exception raised when trying to load that
          entry.  When check=False, metadata retrieval errors will
          raise some kind of error.  When check=True, those are caught
          and returned to the caller.

        """
        # Augmented request -- note that dets:* restrictions from
        # request will be added back into this by check tags.
        aug_request = _filter_items('obs:', request, False)

        if on_missing is None:
            on_missing = {}

        if self.obsdb is not None and 'obs:obs_id' in request:
            if dest is None:
                dest = core.AxisManager()
            obs_man = core.AxisManager()
            obs_info = self.obsdb.get(request['obs:obs_id'], add_prefix='obs:')
            if obs_info is None:
                logger.warning(
                    f"Observation {request['obs:obs_id']} not found in obsdb; "
                    "trying to proceed anyway. You might have metadata failures.")
                obs_man.wrap('obs_id', request['obs:obs_id'])
            else:
                obs_info.update(aug_request)
                aug_request.update(obs_info)
                for k, v in _filter_items('obs:', obs_info).items():
                    obs_man.wrap(k, v)
            dest.wrap('obs_info', obs_man)

        def reraise(spec, e):
            logger.error(
                f"An error occurred while processing a meta entry:\n\n"
                f"  spec:    {spec}\n\n"
                f"  request: {request}\n\n")
            if isinstance(e, LoaderError):
                # Present all args to logger instead...
                for a in e.args[1:]:
                    logger.error(a)
                e = LoaderError(e.args[0])
            raise e

        def check_tags(det_info, aug_request, final=False):
            mask = np.ones(len(det_info), bool)
            unmatched = list(free_tags)
            for tag in free_tags:
                for field in free_tag_fields:
                    if field in det_info.keys:
                        s = (det_info[field] == tag)
                        if s.any():
                            mask *= s
                            unmatched.remove(tag)
            if final and len(unmatched):
                raise RuntimeError(
                    f'One or more free tags was left unconsumed: {unmatched}')

            det_reqs = _filter_items('dets:', request, True)
            unmatched = []
            for k, v in det_reqs.items():
                if k in det_info.keys:
                    if isinstance(v, (list, np.ndarray)):
                        mask *= (core.util.get_multi_index(v, det_info[k]) >= 0)
                    else:
                        mask *= (det_info[k] == v)
                        aug_request['dets:' + k] = v
                else:
                    unmatched.append('dets:' + k)
            if final and len(unmatched):
                raise RuntimeError(
                    f'One or more dets:* selections was left unconsumed: {unmatched}')

            if not np.all(mask):
                logger.debug(f' ... free tags / request reduce det_info (row count '
                             f'{len(det_info)} -> {mask.sum()})')
                det_info = det_info.subset(rows=mask)

            if len(mask) > 0 and len(det_info) == 0:
                logger.warning(f'All detectors have been eliminated from processing.')
                logger.warning(f'  dets:*: {det_reqs}')
                logger.warning(f'  free_tags: {free_tags}')

            return det_info, aug_request

        det_info, aug_request = check_tags(det_info, aug_request)
        n_dets = len(det_info)

        # Process each item.
        items = []
        for spec in spec_list:
            spec, _spec = MetadataSpec.from_dict(spec), spec
            if det_info_scan and not spec.det_info:
                continue

            logger.debug(f'Processing metadata spec={_spec} with augmented '
                         f'request={aug_request}')

            label = spec.label
            _on_missing = spec.on_missing
            if label is not None and label in on_missing:
                _on_missing = on_missing[label]
                logger.debug(f'User overrides on_missing={_on_missing} for {label}')

            assert _on_missing in ['trim', 'skip', 'fail']

            try:
                item = self.load_one(spec, aug_request, det_info)
                error = None
            except Exception as e:
                if check:
                    error = e
                elif ignore_missing or _on_missing == 'skip':
                    logger.warning(f'Failed to load metadata for spec={_spec}; ignoring.')
                    continue
                else:
                    reraise(_spec, e)

            if spec.det_info and error is None:
                item_keys = _filter_items('dets:', item.keys)
                try:
                    det_info = merge_det_info(
                        det_info, item,
                        on_missing=_on_missing)
                except IncompleteMetadataError as e:
                    if check:
                        # I guess we report this, either way.
                        error = e
                    elif _on_missing == 'fail':
                        reraise(_spec, e)
                    elif _on_missing == 'skip':
                        # print a warning I guess
                        logger.warning(f'Skipping failed det_info load, spec={_spec}')

                item = None

                # The check_tags call can cause truncation of the
                # dataset, and that's ok.
                det_info, aug_request = check_tags(det_info, aug_request)
                n_dets = len(det_info)

            if check:
                items.append((spec, error))
                continue

            if item is None:
                # Exit for the det_info case.
                continue

            # Make everything an axisman.
            if isinstance(item, ResultSet):
                # Note this might raise an IncompleteDetInfoError.
                item = broadcast_resultset(item, det_info=det_info)

            elif not isinstance(item, core.AxisManager):
                logger.error(
                    f'The decoded item {item} is not an AxisManager or '
                    f'other well-understood type.  Request was: {request}.')

            if 'dets' in item:
                # You have to check for detector loss here -- compare
                # item.dets.vals to what's in det_info.
                i0 = core.util.get_multi_index(
                    item.dets.vals, det_info['readout_id'])

                n_dets_item = len(set(i0[i0>=0]))
                if n_dets_item < n_dets:
                    message = (f"Only {n_dets_item} of {n_dets} detectors "
                               "have data for metadata specified by "
                               f"spec={_spec}. ")
                    if _on_missing == 'trim':
                        logger.warning(message + 'Trimming.')
                    elif _on_missing == 'fail':
                        raise IncompleteMetadataError(message)
                    else:  # skip
                        logger.warning(message + 'Discarding.')
                        continue

            # Unpack it.
            try:
                dest = unpack_item(spec.unpack, item, dest=dest)
            except Exception as e:
                reraise(_spec, e)

            if 'dets' in dest:
                logger.debug(f'load(): dest now has shape {dest.shape}')
                n_dets = dest.dets.count

        check_tags(det_info, aug_request, final=True)

        if check:
            return items

        dest.wrap('det_info', convert_det_info(det_info))

        return dest


def _filter_items(prefix, d, remove=True):
    # Restrict d to only items that start with prefix; if d is a dict,
    # return a dict with only the keys that satisfy that condition.
    if isinstance(d, dict):
        return {k: d[prefix*remove + k]
                for k in _filter_items(prefix, list(d.keys()), remove=remove)}
    return [k[len(prefix)*remove:] for k in d if k.startswith(prefix)]


def merge_det_info(det_info, new_info, multi=True, on_missing='trim'):
    """Args:

      det_info (ResultSet or None): The det_info table to start from,
        with columns *without* the 'dets:' prefix.
      new_info (ResultSet): New data to merge/check against
        det_info; only columns with dets: prefix are processed.
      multi (bool): whether to permit some rows to match multiple
        rows; this is True by default.
      on_missing (str): what to do if the new_info does not fully cover
        the dets in det_info.

    Returns:
      A (possibly) new det_info table, containing updates and
      restrictions from new_info.  This will contain at least the
      columns that det_info had, and at most as many rows.

    Notes:
      The input arguments det_info and new_info may be modified by
      this function.  Passing in None for det_info is convenient to
      initialize a det_info from a new_info where the columns are
      named dets:... .

    """
    new_keys = _filter_items('dets:', new_info.keys)
    if (len(new_keys) != len(new_info.keys)):
        raise RuntimeError(
            f'New det_info metadata has keys without prefix "dets:": '
            f'{new_info}')
    new_info.keys = new_keys

    if det_info is None:
        return new_info

    join_on = list(set(new_info.keys).intersection(det_info.keys))
    if len(join_on) == 0:
        raise IncompleteMetadataError(
            f'Cannot merge det_info: no common keys in '
            f'{det_info} and {new_info}.')

    new_index = list(zip(*[new_info[k] for k in join_on]))
    det_index = list(zip(*[det_info[k] for k in join_on]))

    # Perform the row-row matching, accepting that rows of new_info
    # may match multiple entries of det_info.  (This is so new_info
    # can have a det_id = 'NO_MATCH' entry.)
    i0 = core.util.get_multi_index(
        new_index, det_index)
    i1 = np.arange(len(det_index))
    i0, i1 = i0[i0>=0], i1[i0>=0]

    if not multi and len(i0) != len(set(i0)):
        offenders = collections.Counter(i0)
        offenders = [new_index[i] for i, v in offenders.items() if v > 1][:10]
        raise ValueError(
            'Some new items matched up to more than one entry in the '
            'existing det_info.  Should this entry have multi=True? '
            f'Offenders: {join_on} = {offenders}')

    # Common fields need to be in accordance, then drop them.
    common_keys = set(new_info.keys) & set(det_info.keys)
    for k in common_keys:
        if len(i0) and np.any(new_info[k][i0] != det_info[k][i1]):
            raise ValueError(
                'When reconciling new det_info, a disagreement in the value '
                f'of field {k} was observed.  If this is due to ambiguity '
                f'in the det_id, maybe add {k} to the match_keys?')

    if len(det_info) != len(i1) and on_missing != 'trim':
        raise IncompleteMetadataError(
            f"det_info had n_dets={len(det_info)}; After merge it has n_dets={len(i1)}"
        )

    logger.debug(f' ... updating det_info (row count '
                 f'{len(det_info)} -> {len(i1)})')
    det_info = det_info.subset(rows=i1)
    new_info = new_info.subset(
        [k for k in new_info.keys if k not in common_keys],
        rows=i0)
    det_info.merge(new_info)
    return det_info


def convert_det_info(det_info, dets=None):
    """Convert det_info ResultSet into nested AxisManager.

    Args:
      dets (list of str): The labels to use for the LabelAxis.  You
        probably want to use the default, det_info['readout_id'], or
        pass in det_info[something_else].

    Returns:

      Nested AxisManager with a LabelAxis called dets, containing all
      the columns from det_info.  Keys in det_info are split on '.';
      so for example det_info['sky.x'] will show up at output.sky.x,
      a.k.a. output['sky']['x'].

    """
    children = {}
    if dets is None:
        dets = det_info['readout_id']
    output = core.AxisManager(core.LabelAxis('dets', dets))
    subtables = {}
    for k in det_info.keys:
        if '.' in k:
            prefix, subkey = k.split('.', 1)
            if not prefix in subtables:
                subtables[prefix] = []
            subtables[prefix].append(subkey)
        else:
            output.wrap(k, det_info[k], [(0, 'dets')])
    for subtable, subkeys in subtables.items():
        sub_info = det_info.subset(keys=[f'{subtable}.{k}' for k in subkeys])
        sub_info.keys = subkeys
        child = convert_det_info(sub_info, dets)
        output.wrap(subtable, child)
    return output

def unconvert_det_info(aman):
    """Convert a det_info-style AxisManager (back) into a ResultSet... the
    opposite of convert_det_info.

    """
    def get_cols(aman, prefix=''):
        columns = []
        for k, v in aman._fields.items():
            if isinstance(v, core.AxisManager):
                columns.extend(get_cols(v, prefix=k + '.'))
            else:
                columns.append((prefix + k, v))
        return columns
    keys, columns = zip(*get_cols(aman))
    return ResultSet(keys, zip(*columns))


def broadcast_resultset(
        rs, det_info, axis_name='dets', axis_key='readout_id',
        prefix='dets:'):
    """Convert rs from a ResultSet into an AxisManager, but reconciling
    against the det_info table to make sure the output's .dets axis
    matches an indexing column of det_info (such as readout_id).

    Args:
      rs (ResultSet): target to convert.  Each column will become a
        vector in the output, unless it begins with 'dets:' in which
        case it will be reconciled and/or discarded (see Notes).
      det_info (ResultSet): table of detector info to use for
        broadcasting rs onto the output axis.
      axis_name (str): Name to use for the LabelAxis of the output.
      axis_key (str): Name of the column of rs that should be used for
        the values in the LabelAxis of the output.
      prefix (str): The prefix that should be used to identify
        indexing columns in rs.

    Returns:
      AxisManager.

    Notes:
      This function is to be applied to specially formatted ResultSet
      tables that contains a mixture of "indexing columns" and "data
      columns".  The "indexing columns" are the ones that start with
      ``prefix``, and are used to associate values in the data columns
      with specific detectors.  For example::

        >>> print(rs)
        ResultSet<[dets:readout_id,a,b], 1280 rows>
        >>> print(det_info)
        ResultSet<[readout_id,x,y,detset], 1280 rows>
        >>> aman = broadcast_resultset(rs, det_info)
        >>> print(aman, aman.dets)
        AxisManager(a[dets], b[dets], dets:LabelAxis(1280))

      In this case the rs column dets:readout_id contains all the same
      values as det_info['readout_id'], and the AxisManager thus also
      contains all 1280 entries.  The .dets axis has the values in the
      order of det_info['readout_id'].

      To broadcast, fields with the stated prefix must also be found
      in det_info.  For example::

        >>> print(rs)
        ResultSet<[dets:band,dets:wafer,abscal], 14 rows>
        >>> print(det_info)
        ResultSet<[readout_id,band,wafer,detset], 1760 rows>
        >>> aman = broadcast_resultset(rs, det_info)
        >>> print(aman)
        AxisManager(abscal[dets], dets:LabelAxis(1738))

      To use database terminology, the rs table is joined with the
      det_info table on the fields prefixed with ``prefix`` (in this
      case, 'band' and 'wafer').  The remaining columns from rs (just
      'abscal') are populated as fields in the output.  Note how the
      output .dets axis has fewer elements than det_info -- the 14
      rows of rs specifying values for different (band, wafer)
      combinations were enough to match only 1738 of the rows in
      det_info.  (This can happen even when not broadcasting.)

    """
    from sotodlib import core

    # Store short names for each index column.
    index_cols = {}
    for k in rs.keys:
        if k.startswith(prefix):
            index_cols[k] = k[len(prefix):]

    # Construct a map that takes a (tuple of dets:* values) to
    # specific row index of rs.
    row_map = {}
    for i, row in enumerate(rs):
        key = tuple([row[k] for k in index_cols.keys()])
        if key in row_map:
            raise ValueError("Duplicate entries for combined unique key.")
        row_map[key] = i

    # Get index of rs that corresponds to each row in det_info.
    missing_keys = [k for k in index_cols.values()
                    if k not in det_info.keys]
    if len(missing_keys):
        raise IncompleteDetInfoError(
            f"Loaded metadata requires det_info keys {missing_keys}.")

    indices = np.array(
        [row_map.get(tuple(row.values()), -1)
         for row in det_info.subset(keys=index_cols.values())],
        dtype=int)
    mask = (indices >= 0)
    dets = det_info[axis_key][mask]
    indices = indices[mask]  # drop any det_info items not matched

    # Re-order and wrap each field.
    aman = core.AxisManager(core.LabelAxis(axis_name, dets))
    for k in rs.keys:
        if not k.startswith(prefix):
            aman.wrap(k, rs[k][indices], [(0, axis_name)])
    return aman


class MetadataSpec:
    """Container for the canonical metadata specification.

    When constructed from a dict, the following attributes are set
    directly from the corresponding key:

    ``db`` (str, ManifestDb)
        The path to a ManifestDb file. For testing and other purposes,
        this may be passed as a ManifestDb object.  Defaults to None.

    ``det_info`` (bool)
        If True, treat the metadata as a contribution to
        det_info. The metadata will be merged into the active
        det_info object.  Defaults to False.

    ``label`` (str)
        A short string describing the metadata.  This is
        used to target the entry when overriding the default
        ``on_missing`` behavior.  Defaults to None.

    ``loader`` (str)
        The name of the loader class to use when loading the
        data.  This will take precedence over what is
        specified in the ManifestDb, and is normally
        unnecessary but can be used for debugging /
        work-arounds.  Defaults to None.

    ``on_missing`` (str)
        String describing how to proceed in the event that the
        metadata is incomplete (or missing entirely) for the target
        Observation.  The value should one of 'trim', 'skip', or
        'fail'.  Defaults to 'trim'.

    ``unpack`` (list of str)
        Instructions for how to populate the destination AxisManager
        with fields found in this metadata item.  See notes below.

    ``load_fields`` (list of str or None)
        List of fields to load. For AxisManager metadata this may
        include entire child AxisManagers, or fields within them
        using "." for hierarchical addressing. For ResultSet metadata,
        subitems of this list can be specified as dictionaries
        (original_fieldname -> new_fieldname) which allow the user to
        change the fieldname of the subitem. Default is None, which
        means to load all fields. Wildcards are not supported.

    ``drop_fields`` (list of str)
        List of fields (which may contain wildcard character ``*``) to
        drop prior to merging.  Only processed for AxisManager
        metadata.  (The dropping is applied after any restrictions on
        the loading using load_fields).

    The following dict keys are deprecated, but are processed for
    backwards compatibility.

    ``name`` (str or list of str)
        (Deprecated.)  This has been renamed as "unpack", and will be
        copied into that attribute if unpack is not otherwise set.


    Notes
    -----

    In the ``unpack`` list, each must be in one of 4 possible forms,
    shown below, to the left of the ``:``.  The resulting assignment
    operation is shown to the right of the ``:``.  ::

            'dest_name&source_name'  : dest[dest_name] = source[source_name]
            'dest_name&'             : dest[dest_name] = source[dest_name]
            'dest_name&*'            : dest[dest_name] = source[wildcard[0]]
            'dest_name'              : dest[dest_name] = source

    The first three forms cause a single field to be extracted.  The
    3rd form is used to extract a single field and rename it, assuming
    that name is the only one in source (it is an error otherwise).

    The ``unpack`` list may include multiple single field extraction
    entries, but each source field may only be referenced once. (So
    for example ``['a&a', 'b&b']`` is valid but ``['a&a, 'b&a']`` is
    not).

    The fourth form causes the entire item to be merged into the
    target at dest_name.  This can operate alongside any number of
    individual field extractions.

    Examples
    --------

    Here is an example ``context.yaml`` metadata list, showing some
    common formations::

      metadata:
        # assignment
        - label: assignment
          db: '{metadata_dir}/det_match/satp1_det_match_240220m/assignment.sqlite'
          det_info: true
          on_missing: fail
        # focal_plane
        - label: focal_plane
          db: '{manifestdir}/focal_plane/satp1_focal_plane_240308r1/db.sqlite'
          unpack: focal_plane
          on_missing: trim
        # hwp_angles
        - label: hwp_angles
          db: '{manifestdir}/hwp_angles/satp1_hwp_angles_240301m/hwp_angle.sqlite'
          load_fields:
          - hwp_angle_enc1
          - hwp_flags
          unpack:
          - 'hwp_angle&hwp_angle_enc1'
          - '&hwp_flags'
          on_missing: drop
        # starcam
        - label: starcam
          db: '{manifestdir}/starcam_solutions/starcam_solutions_240401m/db.sqlite'
          drop_fields: 'image_data_*'
          unpack: starcam
          on_missing: drop

    Note that all entries have ``label`` and ``db`` elements.  The
    ``label`` is unique (this is not required however).  The paths for
    ``db`` all include ``{manifestdir}``.  This will be replaced by
    the value assigned to ``manifestdir`` in the ``tags`` section of
    the context.yaml file.  Referring to particular entries, by label:

    1. The "assignment" entry declares itself as "det_info: true".
       Thus, it does not have an "unpack" key.  The data will unpack
       as a simple table and be merged into the observation's
       "det_info".  Because "on_missing: fail", it is an error if this
       product can not be fully reconciled against an observation
       without dropping detectors.
    2. The "focal_plane" entry specifies "unpack: focal_plane", which
       means that the entire loaded metadata will be placed into a
       child AxisManager called "focal_plane".  However "on_missing:
       trim" means that the focal_plane result does not need to be
       defined for all detectors.  If any are missing, then all data
       for those dets will be dropped from the loaded observation.
    3. The "hwp_angles" entry has a "load_fields" key, which will
       restrict what data are actually pulled in from the product on
       disk.  This is used in cases where the on-disk product has a
       lot of data in it that is not needed.  Specifying that only a
       small subset of the data are needed can greatly increase
       metadata construction time.  The value for "unpack" is now a
       list, identifying that the loaded "hwp_flags" data can be
       placed directly into "hwp_flags", while "hwp_angle_enc1" should
       be renamed to simply "hwp_angle".  The use of "on_missing:
       drop" means that if this product is not available for this
       observation, it is ok to simply continue on without it.
    4. The "starcam" entry uses "drop_fields" to discard certain
       fields from the loaded data, prior to merging it into the
       observation metadata AxisManager.  In practice this doesn't
       save much in terms of i/o cost; it's better to use
       "load_fields" to explicitly include the list of things you care
       about.  The drop_fields option is aimed at deleting fields from buggy
       data because they fail to concatenate properly after load.

    """

    db = None
    det_info = False
    label = None
    loader = None
    on_missing = 'trim'
    unpack = None
    load_fields = None
    drop_fields = []

    @classmethod
    def from_dict(cls, spec):
        self = cls()
        # canonical ...
        for k in ['db', 'label', 'unpack', 'det_info', 'loader',
                  'on_missing', 'load_fields', 'drop_fields']:
            if k in spec:
                setattr(self, k, spec[k])
        # "name" used to be unpacking instructions.
        if 'name' in spec:
            name = spec['name']
            if isinstance(name, str):
                name = [name]
            if self.label is None:
                self.label = name[0].split('&')[0]
            if self.unpack is None:
                self.unpack = name
            else:
                # Prefer self.unpack but warn.
                warnings.warn(
                    "metadata spec contains 'unpack' and 'name' entries; "
                    "ignoring the latter (except to set the 'label', maybe).",
                    MetadataSpecWarning)
        # Make sure unpack is non-empty.
        if self.unpack is None:
            self.unpack = ['&']
        elif isinstance(self.unpack, str):
            self.unpack = [self.unpack]
        # Promote load_fields string to a list (but leave None alone).
        if isinstance(self.load_fields, str):
            self.load_fields = [self.load_fields]
        # Make sure drop_fields is a (possibly empty) list.
        if self.drop_fields is None:
            self.drop_fields = []
        elif isinstance(self.drop_fields, str):
            self.drop_fields = [self.drop_fields]

        return self


def unpack_item(unpack, item, dest=None, wildcard=None):
    # Execute the unpacking of item described in unpack (assumed from
    # MetadataSpec.unpack).
    if wildcard is None:
        wildcard = list(item._fields.keys())[:1]

    # Make a plan based on the unpacking list
    instructions = []
    for name in unpack:
        if '&' in name:
            dest_name, src_name = name.split('&') # check count...
            if src_name == '':
                src_name = dest_name
            elif src_name == '*':
                assert(len(wildcard) == 1)
                src_name = wildcard[0]
            instructions.append(('extract', dest_name, src_name))
        else:
            instructions.append(('full', name, None))

    if dest is None:
        dest = core.AxisManager()

    # Based on instructions, we may need multiple copies of this
    # item; one for any "full" extraction, and one to extract data
    # members from.
    copy_count = sum([inst[0] == 'full' for inst in instructions])
    extr_counts = collections.Counter([inst[2] for inst in instructions
                                       if inst[0] == 'extract'])
    extr_max_count = max(extr_counts.values(), default=0)
    assert extr_max_count <= 1  # Multiple extraction of child fields not supported.
    copy_count += extr_max_count

    # Start with the full copies.
    for inst, dest_name, src_name in instructions:
        if inst != 'full':
            continue
        assert src_name is None
        copy_count -= 1
        _item = item
        if copy_count > 0:
            _item = item.copy()
        dest.wrap(dest_name, _item)

    if extr_max_count == 0:
        return dest

    # And now the partial copies.  By assertion, each field is
    # extracted at most once.
    fields_to_delete = list(item._fields.keys())

    for inst, dest_name, src_name in instructions:
        if inst != 'extract':
            continue
        fields_to_delete.remove(src_name)
        if src_name != dest_name:
            item.move(src_name, dest_name)

    for f in fields_to_delete:
        del item[f]

    dest.merge(item)
    return dest


class LoaderInterface:
    """Base class for "Loader" classes.  Subclasses must define, at least,
    the from_loadspec method.

    """
    def __init__(self, detdb=None, obsdb=None):
        """Args:
          detdb (DetDb): db against which to reconcile dets: index data.
          obsdb (ObsDb): db against which to reconcile obs: index data.

        References to the input database are cached for later use.

        """
        #self.detdb = detdb
        #self.obsdb = obsdb

    def from_loadspec(self, load_params, **kwargs):
        """Retrieve a metadata result.

        Arguments:
          load_params: an index dictionary.

        Returns:
          A ResultSet or similar metadata object.

        """
        raise NotImplementedError

    def batch_from_loadspec(self, load_params, **kwargs):
        """Retrieves a batch of metadata results.  load_params should be a
        list of valid index data specifications.  Returns a list of
        objects, corresponding to the elements of load_params.

        The default implementation simply calls self.from_loadspec
        repeatedly; but subclasses are free to do something more optimized.

        """
        return [self.from_loadspec(p, **kwargs) for p in load_params]


def load_metadata(tod, spec, unpack=False):
    """Process a metadata entry for an AxisManager.

    Args:

      tod (AxisManager): The data structure from which to source any
        obs_info and det_info that are needed to process the metadata
        specification.  This
      spec (dict): a metadata specification, such as one might find as
        an element of the "metadata" list in a context.yaml file.
      unpack (bool): if True, and if the spec does not identify as
        det_info, try to unpack the result into an AxisManager and
        return that.  This will result in broadcasting of items
        indexed by det_info fields into a full .dets axis.

    Returns:
      The loaded metadata item, which could be an AxisManager or
      ResultSet.  In the AxisManager case, the axes have likely not
      been resolved against the provided `tod`, so the sample count
      and detector count / ordering may be different.  (The caller can
      merge after the fact.)

    Notes:
      The ``tod`` container needs to contain ``obs_info`` and
      ``det_info`` (including the ``dets`` axis), in order to follow
      any branching instructions for loading the metadata.  This would
      normally be an AxisManager returned by ``Context.get_obs()`` or
      ``get_meta()``.

    """
    loader = SuperLoader()
    det_info = unconvert_det_info(tod.det_info)
    request = {}
    for k, v in tod.obs_info._fields.items():
        request[f'obs:{k}'] = v
    spec = MetadataSpec.from_dict(spec)
    item = loader.load_one(spec, request, det_info)
    if not unpack or spec.det_info:
        return item

    if isinstance(item, ResultSet):
        # Note this might raise an IncompleteDetInfoError.
        item = broadcast_resultset(item, det_info=det_info)

    return unpack_item(spec.unpack, item)



# Copyright (c) 2020-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Tools for loading Level-3 Book format data.

"""

import os
import sys
from datetime import datetime, timezone
import re
import glob

import yaml
import numpy as np
from astropy import units as u
from astropy.table import Table, QTable
import h5py

# Import so3g before any other packages that import spt3g
import so3g
from spt3g import core as c3g

import toast
import toast.spt3g as t3g
import toast.qarray as qa
from toast.observation import default_values as defaults
from toast.instrument import Telescope, Focalplane, Session, GroundSite
from toast.utils import Logger

from ...core.hardware import parse_readout_id
from ...sim_hardware import telescope_tube_wafer
from ..instrument import (
    SOSite,
    SOFocalplane,
    update_creation_time,
    get_tele_wafer_band_name,
)


def parse_book_name(name):
    """Parse an observation book name.

    This returns the book timestamp, telescope, tube, and list of wafers / streams.

    Args:
        name (str):  The name of the book.

    Returns:
        (tuple):  The timestamp, telescope, tube, and wafers.

    """
    log = toast.utils.Logger.get()

    book_pat = re.compile(r"obs_(\d+)_(.*)_(\d+)")
    book_mat = book_pat.match(name)
    if book_mat is None:
        raise RuntimeError(f"'{name}' is not an observation book")
    book_timestamp = float(book_mat.group(1))
    book_teltube = book_mat.group(2)
    book_slots = book_mat.group(3)

    lat_pat = re.compile(r"lat(.*)")
    lat_mat = lat_pat.match(book_teltube)
    if lat_mat is None:
        # This is a sat
        tele_name = book_teltube.upper()
        tube_name = f"ST{tele_name[-1]}"
    else:
        tele_name = "LAT"
        tube_name = lat_mat.group(1)

    ttw = telescope_tube_wafer()
    wafers = list()
    for bit, wf in zip(book_slots, ttw[tele_name][tube_name]):
        if bit == "1":
            wafers.append(wf)

    return book_timestamp, tele_name, tube_name, wafers


def parse_book_time(dstr):
    """Parse a book observation timestamp.

    Args:
        dstr (str):  The timestamp string.

    Returns:
        (datetime):  The datetime object.

    """
    return datetime.strptime(dstr, "%Y-%m-%dT%H:%M:%S%z")


class MetaCollector(object):
    """Class to parse frames and build metadata."""

    def __init__(self):
        self.frame_sizes = list()
        self.readout_ids = None
        self.sample_rate = list()

    def __call__(self, frame):
        if frame is not None and frame.type != c3g.G3FrameType.EndProcessing:
            if frame.type == c3g.G3FrameType.Scan:
                # We have a Scan frame, find the number of samples
                srange = np.array(frame["sample_range"])
                if len(srange) != 2:
                    raise RuntimeError(
                        f"frame sample range has {len(srange)} elements instead of 2"
                    )
                if srange[0] >= srange[1]:
                    raise RuntimeError(
                        f"frame has no samples (range = {srange[0]} - {srange[1]}"
                    )
                self.frame_sizes.append(srange[1] - srange[0])

                # If this is the first Scan frame, get the list of detector readout IDs
                if self.readout_ids is None:
                    self.readout_ids = list(frame["signal"].names)
                    # print(f"metacollect first frame, set readout_ids to {self.readout_ids[0]} .. {self.readout_ids[-1]}", flush=True)

                # Estimate the sample rate from the timestamps
                stamps = t3g.from_g3_time(frame["ancil"].times)
                (rate, dt, dt_min, dt_max, dt_std) = toast.utils.rate_from_times(stamps)
                self.sample_rate.append(rate)
        return


def parse_frame_meta(frame_files):
    """Parse the frame information.

    Args:
        frame_files (list):  List of frame files for one wafer.

    Returns:
        (tuple):  The frame sizes, readout names, sample rate.

    """
    meta_collect = MetaCollector()
    for ffile in frame_files:
        # print(f"running metacollect pipe on {ffile}")
        load_pipe = c3g.G3Pipeline()
        load_pipe.Add(c3g.G3Reader(ffile))
        load_pipe.Add(meta_collect)
        load_pipe.Run()
    return (
        np.array(meta_collect.frame_sizes, dtype=np.int32),
        list(meta_collect.readout_ids),
        np.median(meta_collect.sample_rate),
    )


class DataSpreader(object):
    """Class to distribute frames.

    This class receives frames on the rank zero process of a group and uses
    point-to-point communication to send frames to the correct process.  Because
    This only accumulates and sends one process worth of data at a time, the
    entire observation never exists on the root process.  The tradeoff however
    is a performance hit compared to gathering all data and performing a scatterv.

    """

    def __init__(
        self,
        obs,
        obs_fields,
    ):
        self.obs = obs
        self.obs_fields = obs_fields
        self.dist = obs.dist
        # The distribution sample sets represent the frame sizes.  Make a mapping of
        # frame index to process.
        foff = 0
        self.frame_target = dict()
        for iproc, prange in enumerate(self.dist.samp_sets):
            for ifr in range(prange.n_elem):
                # Each sample set contains exactly one frame
                self.frame_target[foff] = iproc
                foff += 1
        self.cur_frame = 0
        self.cur_proc = 0
        self.send_buffer = list()
        self.recv_buffer = list()

    def __call__(self, frame):
        # This function will only be called on rank 0
        if frame is not None and frame.type == c3g.G3FrameType.Scan:
            # Accumulate scan frame to send buffer
            if self.frame_target[self.cur_frame] != self.cur_proc:
                # Done with this process
                if self.cur_proc == 0:
                    # Copy to our own receive buffer
                    self.recv_buffer = [x for x in self.send_buffer]
                else:
                    self.obs.comm.comm_group.send(
                        self.send_buffer, self.cur_proc, tag=self.cur_proc
                    )
                self.cur_proc = self.frame_target[self.cur_frame]
                self.send_buffer.clear()
            self.send_buffer.append(frame)
            self.cur_frame += 1

    def flush(self):
        if self.obs.comm.group_rank != 0:
            # Wait to receive our list of frames.  This code only runs if the group
            # has more than one process, so no need to check comm for None.
            self.recv_buffer = self.obs.comm.comm_group.recv(
                source=0,
                tag=self.obs.comm.group_rank,
            )
            self._unpack_frames(self.recv_buffer)
        else:
            # Root process flushes remaining buffer
            if self.cur_proc == 0:
                # This means there was only one process- just unpack the
                # send buffer.
                self._unpack_frames(self.send_buffer)
            else:
                # We were accumulating the buffer for another process when
                # the frame stream ended.  Send it now.
                self.obs.comm.comm_group.send(
                    self.send_buffer, self.cur_proc, tag=self.cur_proc
                )
                # Also unpack our own frames
                self._unpack_frames(self.recv_buffer)
        # Reset counters
        self.cur_frame = 0
        self.cur_proc = 0
        self.recv_buffer = list()
        self.send_buffer = list()

    def _unpack_frames(self, buffer):
        # Since the data is distributed in the sample direction, every process has
        # exclusive use of the column shared objects.  This means it is safe to
        # set those buffers directly.
        off = 0
        detmap = None
        # print(f"UNPACK frames for telescope {self.obs.telescope.name}", flush=True)
        fpdata = self.obs.telescope.focalplane.detector_data
        # print(f"  FPDATA = {fpdata[:3]}", flush=True)
        for frm in buffer:
            ancil = frm["ancil"]
            times = t3g.from_g3_time(ancil.times)
            n = len(times)
            self.obs.shared[self.obs_fields["times"]].data[off : off + n] = times
            self.obs.shared[self.obs_fields["azimuth"]].data[off : off + n] = np.array(
                ancil["az_enc"], copy=False
            )
            self.obs.shared[self.obs_fields["elevation"]].data[off : off + n] = ancil[
                "el_enc"
            ]
            if "flags" in ancil:
                self.obs.shared[self.obs_fields["shared_flags"]].data[
                    off : off + n
                ] = np.array(ancil["flags"]).astype(np.uint8)

            self.obs.shared[self.obs_fields["boresight_azel"]].data[
                off : off + n, :
            ] = t3g.from_g3_quats(frm["qboresight_azel"])

            if "qboresight_radec" in frm:
                self.obs.shared[self.obs_fields["boresight_radec"]].data[
                    off : off + n, :
                ] = t3g.from_g3_quats(frm["qboresight_radec"])

            if "hwp_enc" in ancil:
                self.obs.shared[self.obs_fields["hwp_angle"]].data[
                    off : off + n
                ] = np.array(ancil["hwp_enc"]).astype(np.float64)

            if "corotation_enc" in ancil:
                self.obs.shared[self.obs_fields["corotator_angle"]].data[
                    off : off + n
                ] = np.array(ancil["corotation_enc"]).astype(np.float64)

            if "boresight_enc" in ancil:
                self.obs.shared[self.obs_fields["boresight_angle"]].data[
                    off : off + n
                ] = np.array(ancil["boresight_enc"]).astype(np.float64)

            if detmap is None:
                # Compute the mapping between observation detectors and rows of the
                # supertimestream.
                # print(f"  First frame, signal names = {frm['signal'].names[0]} .. {frm['signal'].names[-1]}")
                super_name_to_row = {y: x for x, y in enumerate(frm["signal"].names)}
                # print(super_name_to_row, flush=True)
                detmap = np.array([super_name_to_row[x] for x in fpdata["readout_id"]])
                # print(f"{self.obs.name} detmap = {detmap}")
            # print(f"DBG LOAD frame {off}:{off+n} signal = {frm['signal'].data[:5,:]}")
            # print(f"LOAD frame {off}:{off+n} signal = {frm['signal'].data[detmap,:]}")
            for idet in range(len(detmap)):
                self.obs.detdata[self.obs_fields["det_data"]][
                    idet, off : off + n
                ] = frm["signal"].data[detmap[idet], :]
                self.obs.detdata[self.obs_fields["det_flags"]][
                    idet, off : off + n
                ] = np.array(frm["flags"].data[detmap[idet], :], dtype=np.uint8)
            if "signal_units" in frm and frm["signal_units"] is not None:
                # We have some unit information
                # print(f"Frame units = {frm['signal_units']}")
                self.obs.detdata[self.obs_fields["det_data"]].update_units(
                    t3g.from_g3_unit(frm["signal_units"])
                )
            off += n


def import_obs_data(
    frame_files,
    ob,
    fields,
):
    # Create observation fields
    ob.shared.create_column(
        fields["times"],
        shape=(ob.n_local_samples,),
        dtype=np.float64,
    )
    ob.shared.create_column(
        fields["azimuth"],
        shape=(ob.n_local_samples,),
        dtype=np.float64,
    )
    ob.shared.create_column(
        fields["elevation"],
        shape=(ob.n_local_samples,),
        dtype=np.float64,
    )
    ob.shared.create_column(
        fields["boresight_azel"],
        shape=(ob.n_local_samples, 4),
        dtype=np.float64,
    )
    ob.shared.create_column(
        fields["boresight_radec"],
        shape=(ob.n_local_samples, 4),
        dtype=np.float64,
    )
    if fields["hwp_angle"] is not None:
        ob.shared.create_column(
            fields["hwp_angle"],
            shape=(ob.n_local_samples,),
            dtype=np.float64,
        )
    if fields["boresight_angle"] is not None:
        ob.shared.create_column(
            fields["boresight_angle"],
            shape=(ob.n_local_samples,),
            dtype=np.float64,
        )
    if fields["corotator_angle"] is not None:
        ob.shared.create_column(
            fields["corotator_angle"],
            shape=(ob.n_local_samples,),
            dtype=np.float64,
        )
    if fields["shared_flags"] is not None:
        ob.shared.create_column(
            fields["shared_flags"],
            shape=(ob.n_local_samples,),
            dtype=np.uint8,
        )

    ob.shared.create_column(
        defaults.position,
        shape=(ob.n_local_samples, 3),
        dtype=np.float64,
    )
    ob.shared.create_column(
        defaults.velocity,
        shape=(ob.n_local_samples, 3),
        dtype=np.float64,
    )

    ob.detdata.create(fields["det_data"], dtype=np.float64)
    ob.detdata.create(fields["det_flags"], dtype=np.uint8)

    # Distribute frames and copy into place

    spreader = DataSpreader(ob, fields)
    if ob.comm.group_rank == 0:
        # Read and send
        for ffile in frame_files:
            load_pipe = c3g.G3Pipeline()
            load_pipe.Add(c3g.G3Reader(ffile))
            load_pipe.Add(spreader)
            load_pipe.Run()
    # Unpack local frames
    spreader.flush()


@toast.timing.function_timer
def read_book(
    comm,
    fp_nominal,
    book_dir,
    focalplane_dir,
    noise_dir,
    obs_fields,
    tube,
    wafer_bands,
    detset_key=None,
    frame_intervals=None,
    ignore_sim=False,
):
    """Read a book into single observing session.

    Observations are created for each selected wafer and band.

    Args:
        comm (toast.Comm):  The toast communicator.
        fp_nominal (SOFocalplane):  The nominal focalplane.
        book_dir (str):  The directory containing the book.
        focalplane_dir (str):  Path to directory of focalplane files.
        noise_dir (str):  Path to directory of noise models.
        obs_fields (dict):  Dictionary of observation fields to use.
        tube (str):  The tube slot.
        wafer_bands (dict):  Dictionary of bands to load for each wafer.
        frame_intervals (str):  The intervals to use for frame boundaries.
        ignore_sim (bool):  If True, do not load toast simulation information
            in the M_index.yaml file.  This will emulate loading real data,
            but the resulting observation cannot be used for reproducing the
            simulated content.

    Returns:
        (list):  The list of observations for this session.

    """
    log = Logger.get()
    book_name = os.path.basename(book_dir)
    # One process loads book metadata
    book_meta = None
    obs_meta = None
    if comm.group_rank == 0:
        book_meta_file = os.path.join(book_dir, f"M_book.yaml")
        # print(f"Loading {book_meta_file}")
        with open(book_meta_file, "r") as f:
            book_meta = yaml.load(f, Loader=yaml.Loader)
        # print(f"{book_dir}:  {book_meta}")
        obs_meta_file = os.path.join(book_dir, f"M_index.yaml")
        # print(f"Loading {obs_meta_file}")
        with open(obs_meta_file, "r") as f:
            obs_meta = yaml.load(f, Loader=yaml.Loader)
        # print(f"{book_dir}:  {obs_meta}")
    if comm.comm_group is not None:
        book_meta = comm.comm_group.bcast(book_meta, root=0)
        obs_meta = comm.comm_group.bcast(obs_meta, root=0)

    # All of the observation metadata is currently duplicated between the observation
    # frame and the M_index.yaml file.  However, if that changes we could grab
    # the observation frame from any of the files here.

    book_session = False
    tele_uid = None
    if not ignore_sim:
        # Attempt to restore the session and site information that was used during
        # the sim.
        toast_meta_file = os.path.join(book_dir, f"M_toast.yaml")
        if os.path.isfile(toast_meta_file):
            with open(toast_meta_file, "r") as f:
                toast_meta = yaml.load(f, Loader=yaml.Loader)
            try:
                tele_uid = toast_meta["telescope_uid"]
            except KeyError:
                msg = f"Book {book_name} toast sim info does not have telescope UID "
                msg += f"Will use default value based on telescope name."
                log.warning_rank(msg, comm=comm.comm_group)
            try:
                sname = toast_meta["observing_session"]
                sstart = toast_meta["observing_session_start"]
                send = toast_meta["observing_session_end"]
                suid = None
                if "observing_session_uid" in toast_meta:
                    suid = toast_meta["observing_session_uid"]
                session = Session(
                    sname,
                    uid=suid,
                    start=datetime.fromtimestamp(sstart).astimezone(timezone.utc),
                    end=datetime.fromtimestamp(send).astimezone(timezone.utc),
                )
            except KeyError:
                msg = f"Book {book_name} toast sim info does not have valid observing "
                msg += f" session.  Will use nominal values."
                log.warning_rank(msg, comm=comm.comm_group)
                book_session = True

            weather = None
            try:
                wname = toast_meta["site_weather_name"]
                wreal = toast_meta["site_weather_realization"]
                wtime = toast_meta["site_weather_time"]
                wsuid = toast_meta["site_weather_uid"]
                wmaxpwv = toast_meta["site_weather_max_pwv"]
                if wmaxpwv == "NONE":
                    wmaxpwv = None
                else:
                    wmaxpwv = u.Quantity(wmaxpwv)
                use_median = bool(toast_meta["site_weather_use_median"])
                weather = toast.weather.SimWeather(
                    time=datetime.fromtimestamp(wtime).astimezone(timezone.utc),
                    name=wname,
                    site_uid=wsuid,
                    realization=wreal,
                    max_pwv=wmaxpwv,
                    median_weather=use_median,
                )
            except KeyError:
                # weather info may not exist
                pass

            try:
                sialt = toast_meta["site_alt_m"]
                silat = toast_meta["site_lat_deg"]
                silon = toast_meta["site_lon_deg"]
                siname = toast_meta["site_name"]
                siuid = None
                if "site_uid" in toast_meta:
                    siuid = toast_meta["site_uid"]
                site = GroundSite(
                    siname,
                    silat * u.degree,
                    silon * u.degree,
                    sialt * u.meter,
                    uid=siuid,
                    weather=weather,
                )
            except KeyError:
                msg = f"Book {book_name} toast sim info does not have valid "
                msg += f"site information.  Will use nominal values."
                log.warning_rank(msg, comm=comm.comm_group)
                book_session = True
        else:
            msg = f"Cannot load simulation info from book '{toast_meta_file}', "
            msg += f"use ignore_sim=True?"
            log.warning_rank(msg, comm=comm.comm_group)
    else:
        book_session = True

    if book_session:
        # Create a nominal Site and a Session based only on the book information
        session = Session(
            book_name,
            start=datetime.fromisoformat(obs_meta["timestamp_start"]),
            end=datetime.fromisoformat(obs_meta["timestamp_end"]),
        )
        site = SOSite()

    session_obs = list()

    # Compute the mapping from wafer and channel to row of the nominal focalplane
    wchan = dict()
    for wafer, props in wafer_bands.items():
        wchan[wafer] = dict()
    for irow, row in enumerate(fp_nominal.detector_data):
        if row["wafer_slot"] in wafer_bands:
            wchan[row["wafer_slot"]][row["channel"]] = (irow, row["band"])

    for wafer, props in wafer_bands.items():
        bands = props["bands"]
        framefiles = glob.glob(os.path.join(book_dir, f"D_{wafer}_*.g3"))
        framefiles.extend(glob.glob(os.path.join(book_dir, f"D_{wafer}_*.g3.gz")))

        # Esure that frame files are processed in order!
        framefiles.sort()

        # One process gets the D-frame info
        frame_sizes = None
        readout_ids = None
        sample_rate = None
        if comm.group_rank == 0:
            frame_sizes, readout_ids, sample_rate = parse_frame_meta(framefiles)
        if comm.comm_group is not None:
            frame_sizes = comm.comm_group.bcast(frame_sizes, root=0)
            readout_ids = comm.comm_group.bcast(readout_ids, root=0)
            sample_rate = comm.comm_group.bcast(sample_rate, root=0)
        sample_rate = u.Quantity(sample_rate * u.Hz)
        _, creation_time, _ = parse_readout_id(readout_ids[0])

        # Use the frame boundaries to create sample sets each with one frame
        sample_sets = [[x] for x in frame_sizes]
        total_samples = np.sum(frame_sizes)

        for bnd in bands:
            # Do we have a focalplane model for this data?  If not, we will create a
            # nominal one as a starting point.
            use_nominal = True
            if focalplane_dir is not None:
                fp_file = os.path.join(focalplane_dir, f"{book_name}_{wafer}_{bnd}.h5")
                # print(f"Loading focalplane {fp_file}", flush=True)
                if os.path.isfile(fp_file):
                    # We have a file, load it
                    focalplane = toast.instrument.Focalplane(sample_rate=sample_rate)
                    with toast.io.H5File(
                        fp_file, "r", comm=comm.comm_group, force_serial=True
                    ) as f:
                        focalplane.load_hdf5(f.handle, comm=comm.comm_group)
                    update_creation_time(focalplane.detector_data, creation_time)
                    use_nominal = False
                else:
                    log.warning(
                        f"Focalplane file {fp_file} does not exist, using nominal values"
                    )
            if use_nominal:
                # Find the rows of the frame data which match our band
                fp_rows = list()
                for rid in readout_ids:
                    wf, ct, chan = parse_readout_id(rid)
                    fpr, fpband = wchan[wf][chan]
                    if fpband == bnd:
                        fp_rows.append(fpr)

                # Create a focalplane by extracting these rows from the nominal one.
                detprops = QTable(fp_nominal[fp_rows])
                update_creation_time(detprops, creation_time)
                focalplane = toast.instrument.Focalplane(
                    detector_data=detprops, sample_rate=sample_rate
                )

            # Create the telescope

            tele_name = get_tele_wafer_band_name(
                obs_meta["telescope"],
                tube,
                wafer,
                bnd,
            )
            telescope = Telescope(
                tele_name, uid=tele_uid, focalplane=focalplane, site=site
            )
            obs_name = f"{session.name}_{tele_name}"

            # Create the observation

            if detset_key is None:
                detsets = None
            else:
                detsets = focalplane.detector_groups(detset_key)

            obs = toast.Observation(
                comm,
                telescope,
                total_samples,
                name=obs_name,
                session=session,
                detector_sets=detsets,
                sample_sets=sample_sets,
                process_rows=1,
            )

            # Load frame data into the observation
            # print(f"CALL import obs data:  {framefiles} | {obs_fields}")
            import_obs_data(
                framefiles,
                obs,
                obs_fields,
            )

            # Position and velocity of the observatory are simply computed
            position, velocity = site.position_velocity(obs.shared[obs_fields["times"]])
            obs.shared[defaults.position].data[:] = position
            obs.shared[defaults.velocity].data[:] = velocity

            if noise_dir is not None:
                # FIXME: we could regex match on a glob here to find the
                # observation key name to use.
                noise_file = os.path.join(
                    noise_dir, f"{book_name}_{wafer}_{bnd}_noise_model.h5"
                )
                # print(f"Loading focalplane {fp_file}", flush=True)
                if os.path.isfile(noise_file):
                    # We have a file, load it
                    nse = toast.noise.Noise()
                    with toast.io.H5File(noise_file, "r", comm=comm.comm_group) as f:
                        nse.load_hdf5(f.handle, obs)
                    obs["noise_model"] = nse
                else:
                    log.warning_rank(
                        f"Noise model file {noise_file} does not exist, skipping",
                        comm=comm.comm_group,
                    )

            session_obs.append(obs)

    return session_obs
